#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2017-2020 caregraf

import os
import sys
import re
import json
from collections import defaultdict, Counter
from datetime import datetime, date
import logging

from ..reporter.reportUtils import reportAbsAndPercent

"""
Using data, defines a type according to the presence and (selectively) the frequency distribution of the values of properties. Optionally splits on the values of one or more nominal properties to distinguish finer 'subtypes'.

REM: VistA typing/files/collections n enums etc often too broad, too narrow or muddy.
(Sub)Typing pictures VistA but often needs better categorization/groupings, things
refined from reframing the data first and this utility does allow for that.
  
Arguments:
- subTypeProps: one or more properties used to subtype by. Note that
pointer label is used for pointer properties
- createDateProp: isolates the property that designates the creation date of a record
  ... may change to be 'recordDateProp' as for scheduling, making this the 
  start time of an appt and not the time an appt is created.
- maxNominalValues: for pointer or literals - by default won't be counted once
distinct value count exceeds this threshold. Ensures type reduction doesn't become
a large restatement of the records 
- countDateType: by default only counts dates by year but MONTH, WEEK, DAY are other options
- forceCountPointerTypes: these pointer values are counted even if they exceed the maxNominalValues threshold.
- forceCountProperties: applies to both POINTER and LITERAL (usually enums) properties. Usually just for enums but may be used for variable pointer pointers too
- deidentifyPointerTypes: remove label from these (PATIENT set by default)

Note that there is a default set of FORCE_COUNTED PTER TYPEs covering user (200),
locations and patient.
"""
DEFAULT_MAX_NOMINAL_VALUES = 60 # maxNominalValues=60 as 55 is # dialogs of 100
DEFAULT_FORCE_COUNT_PTER_TYPES = ["4", "44", "200"] # user, location, patient
USER_NAME_TYPES = ["220_5"]
DEFAULT_FORCE_COUNT_PTER_TYPES.extend(USER_NAME_TYPES)
PATIENT_NAME_TYPES = ["2", "9000001", "55", "63", "70", "631", "665", "690", "790", "38_1", "354_7"] # mainly dinumed, 2005 as its label has id info
DEFAULT_FORCE_COUNT_PTER_TYPES.extend(PATIENT_NAME_TYPES)

class SubTypeReducer():
    
    def __init__(self, typId, subTypeProps, createDateProp="", maxNominalValues=DEFAULT_MAX_NOMINAL_VALUES, countDateType="YEAR", forceCountPointerTypes=DEFAULT_FORCE_COUNT_PTER_TYPES, forceCountProperties=[], deidentifyPointerTypes=PATIENT_NAME_TYPES):
    
        self.__typId = typId
        self.__subTypeProps = subTypeProps
        self.__createDateProp = createDateProp
        self.__maxNominalValues = maxNominalValues
        self.__countDateType = countDateType
        self.__forceCountPointerTypes = set(forceCountPointerTypes)
        self.__forceCountProperties = set(forceCountProperties)
        self.__deidentifyPointerTypes = set(deidentifyPointerTypes)
        
        self.__typeReducers = {}
        # minimal for 'all' option as sub types have details
        self.__allTypeReducer = TypeReducer(
            self.__typId, 
            createDateProp=self.__createDateProp, 
            maxNominalValues=self.__maxNominalValues, 
            countDateType="YEAR", # no nuance - subtypes get nuance 
            forceCountPointerTypes=[], 
            forceCountProperties=[],
            deidentifyPointerTypes=self.__deidentifyPointerTypes
        ) 
                    
    def transform(self, resource):
        # for all
        self.__allTypeReducer.transform(resource)  
        
        # will create a (Sub)TypeReducer for any combos seen - REM: pointer label used
        pieces = ["-".join(self.__subTypeProps)]
        for stp in self.__subTypeProps:    
            if re.match(r'\#', stp): # allows presence/absence vs counting
                stpv = "*PRESENT*" if stp.split("#")[1] in resource else "*ABSENT*"
            elif stp not in resource:
                stpv = "**NOVALUE**"
            else:
                if isinstance(resource[stp], list):
                    stpv = "**HASMEMBERS**"
                elif isinstance(resource[stp], dict):
                    if "value" in resource[stp]: # will only split by year
                        stpv = resource[stp]["value"].split("-")[0]
                    elif "label" in resource[stp]:
                        # PY2 -> 3: removed .encode("ascii", "ignore") for label
                        stpv = "{} [{}]".format(resource[stp]["label"], resource[stp]["id"])
                    else: # should never happen
                        raise Exception("Expect structured value of sub type property {} to be a POINTER or DATE (Year)".format(stp))
                elif isinstance(resource[stp], bool):
                    stpv = str(resource[stp])
                elif len(resource[stp].split(":")) == 2: # enum (allows more complex : to stay)
                    stpv = resource[stp].split(":")[1]
                else:
                    stpv = resource[stp]
            pieces.append(re.sub(r' ', '_', stpv))
        subTypeId = ":".join(pieces)
        typeReducer = self.__lookupSubTypeReducer(subTypeId)
        typeReducer.transform(resource)
        
    def subTypeReducerCount(self):
        return len(self.__typeReducers)
        
    def reductions(self):
        results = []
        allReduction = self.__allTypeReducer.reduction()
        allReduction["_reductionEndDay"] = datetime.now().strftime("%Y-%m-%d")
        results.append(allReduction)
        allTotal = allReduction["_total"]
        for subTypeId in self.__typeReducers:
            reduction = self.__typeReducers[subTypeId].reduction()
            reduction["_subTypeId"] = subTypeId
            reduction["_numberFiltered"] = allTotal - reduction["_total"]
            results.append(reduction)
        return results
        
    def __lookupSubTypeReducer(self, subTypeId):
    
        if subTypeId not in self.__typeReducers:
            self.__typeReducers[subTypeId] = TypeReducer(
                self.__typId, 
                createDateProp=self.__createDateProp, 
                maxNominalValues=self.__maxNominalValues, 
                countDateType=self.__countDateType,
                forceCountPointerTypes=self.__forceCountPointerTypes, 
                forceCountProperties=self.__forceCountProperties, 
                deidentifyPointerTypes=self.__deidentifyPointerTypes
            )
            
        return self.__typeReducers[subTypeId]
                            
class TypeReducer():

    def __init__(self, typId, reductionLabel="", createDateProp="", maxNominalValues=DEFAULT_MAX_NOMINAL_VALUES, countDateType="YEAR", forceCountPointerTypes=DEFAULT_FORCE_COUNT_PTER_TYPES, forceCountProperties=[], deidentifyPointerTypes=PATIENT_NAME_TYPES):
        
        self.__typId = typId
        self.__createDateProp = createDateProp
        self.__maxNominalValues = maxNominalValues
        self.__countDateType = countDateType
        self.__forceCountPointerTypes = set(forceCountPointerTypes)
        self.__forceCountProperties = set(forceCountProperties)
        self.__deidentifyPointerTypes = set(deidentifyPointerTypes)

        self.__dontCountLiteralProp = set()
        self.__distinctPointerValuesByProp = defaultdict(lambda: set())
        self.__typeData = {"_total": 0, "_id": typId}
        if reductionLabel: # REM: filter can make a different subset
            self.__typeData["_reductionLabel"] = reductionLabel
        if createDateProp and not callable(createDateProp):
            self.__typeData["_createDateProp"] = createDateProp

    def transform(self, resource):
        self.__reduce(resource, self.__typeData)
        
    def subTypeReducerCount(self):
        return 0
        
    def reductions(self):
        return [self.__typeData]

    def reduction(self):
        return self.__typeData

    """
    Works for both contained resources and resources
    """
    def __reduce(self, resource, typeData, parentPropPath=""):

        def isPointer(value):
            if len(value) == 0:
                return False # shouldn't happen but some synthesized has {} value
            if len(set(value) - set(["id", "sameAs", "sameAsLabel", "label"])):
                return False
            return True
        
        def typeProp(value): # WHAT ABOUT BOOLEAN?
            if isinstance(value, bool):
                return "BOOLEAN"
            if isinstance(value, int):
                return "INTEGER"
            if isinstance(value, dict):
                if "value" in value:
                    return "DATE"
                elif isPointer(value):
                    return "POINTER"
                else: # assuming - and needn't have an 'id'
                    return "OBJECT"
            if isinstance(value, list):
                return "LIST" # TODO - simple list work ie/ not {"x": y ...} but [y ...]
            return "LITERAL"

        # allows for label based date value (ie/ pter to visit etc)
        def getCreateDate(resource, createDateProp):
            if not createDateProp:
                return None
            if callable(createDateProp):
                return createDateProp(resource)
            # Allow for real date and a pointer who's label is a date!
            if createDateProp not in resource:
                return None
            if not isinstance(resource[createDateProp], dict):
                return None
            if not ("value" in resource[createDateProp] or "label" in resource[createDateProp]):
                return None
            if "value" in resource[createDateProp]:
                dtValue = resource[createDateProp]["value"]
            elif "label" in resource[createDateProp]:
                dtValue = resource[createDateProp]["label"]
            if not validDayDT(dtValue.split("T")[0]): 
                return None
            return dtValue
    
        # at start, assume all literals are enums AND handle booleans here too
        def countByLiteralValue(qprop, propInfo, litValue): 
            if qprop in self.__dontCountLiteralProp:
                return
            # First exclusion: if long then a real literal and not just a value 
            # (NOTE: bool's fall in here too)
            if isinstance(litValue, str) and qprop not in self.__forceCountProperties and len(litValue.split(" ")) > 5:
                self.__dontCountLiteralProp.add(qprop)
                if "byValueCount" in propInfo:
                    del propInfo["byValueCount"]
                return
            if "byValueCount" not in propInfo:
                propInfo["byValueCount"] = { litValue: 1 }
                return
            if litValue not in propInfo["byValueCount"]:
                # second exclusion - too many but can force a pass 
                if qprop not in self.__forceCountProperties and len(propInfo["byValueCount"]) == self.__maxNominalValues:
                    self.__dontCountLiteralProp.add(qprop)
                    del propInfo["byValueCount"]
                    return 
                propInfo["byValueCount"][litValue] = 0
            propInfo["byValueCount"][litValue] +=1

        def countByPointerValue(qprop, propInfo, ptrValue):
            """
            Either 
            - count distinct values (if large number) OR
            - count by if:
              - small as acts as enum (ie/ <= enum threshold)
              - forced to count (usually for 200/2)
            - accounts for multiple prop and parent prop having same name by using
            qprop to track size
              
            TODO: 
            - distinct pointer values by prop could get big for self references of
            big types (2005 to 2005 etc). May need cut off there too
            - reconsider vptr with other overwhelming 200 or 4 etc (may never happen)
            """
            sptrValue = ptrValue
            if re.search(r'\[', ptrValue):
                sptrValue = re.search(r'\[([^\]]+)', ptrValue).group(1)
            MAX_RANGE_COUNT = 1000000 # for 2005 to 2005 etc.
            if qprop in self.__distinctPointerValuesByProp:
                if "rangeCount" in propInfo:
                    self.__distinctPointerValuesByProp[qprop].add(sptrValue)
                    if len(self.__distinctPointerValuesByProp[qprop]) == MAX_RANGE_COUNT:
                        logging.debug("Property {} has too big a range - >={:,} - so not even counting that".format(qprop, MAX_RANGE_COUNT))
                        del propInfo["rangeCount"]
                        self.__distinctPointerValuesByProp[qprop] == None
                    else: 
                        propInfo["rangeCount"] = len(self.__distinctPointerValuesByProp[qprop])
                return
            if "byValueCount" not in propInfo:
                propInfo["byValueCount"] = { ptrValue: 1 }
                return
            if ptrValue not in propInfo["byValueCount"]:
                ptrType = sptrValue.split("-")[0]
                """ 
                VPTRs present difficulties - if just a few of forced pointer then will
                fail to count em unless appear in first set
                
                TODO: may renew and simply record those forced - effected 442096
                """
                if not (qprop in self.__forceCountProperties or ptrType in self.__forceCountPointerTypes) and len(propInfo["byValueCount"]) == self.__maxNominalValues:
                    self.__distinctPointerValuesByProp[qprop] = set([
                        re.search(r'\[([^\]]+)', ptrValue).group(1) if re.search(r'\[', ptrValue) else ptrValue for ptrValue in propInfo["byValueCount"]])
                    propInfo["rangeCount"] = len(self.__distinctPointerValuesByProp[qprop])
                    del propInfo["byValueCount"]
                    logging.debug("Property {} is too big for byValueCount - just measure range".format(qprop))
                    return 
                propInfo["byValueCount"][ptrValue] = 0
            propInfo["byValueCount"][ptrValue] +=1
            
        # as opposed to reduction of all sresources, this establishes number per parent
        def countListValueByLength(propInfo, listValue):
            if "byValueCount" not in propInfo:
                propInfo["byValueCount"] = {}
            if len(listValue) in propInfo["byValueCount"]:
                propInfo["byValueCount"][len(listValue)] += 1
                return
            propInfo["byValueCount"][len(listValue)] = 1

        """
        Compromise that gives data for graphing, trending
        time including weekday/weekend splits and per day
        creation graphs but which doesn't collect too much 
        for types with many dates.
        """
        def countDateValue(prop, propInfo, dtValue):
            dayValue = dtValue.split("T")[0]
            dayDT = validDayDT(dayValue)
            if not dayDT:
                return
            dtps = dayValue.split("-")
            monthValue = "{}-{}".format(dtps[0], dtps[1])
            # Not doing WEEK as WEEK in what year?
            #     dtToCount = dayDT.isocalendar()[1]                
            if self.__countDateType == "MONTH":
                dtToCount = monthValue
            elif self.__countDateType == "DAY":
                # May downstep all but createProp no matter what
                # if self.__createDateProp != prop:
                #    dtToCount = monthValue
                # else:
                dtToCount = dayValue
            else: # "YEAR" is default
                dtToCount = dtps[0]            
            if "byValueCount" not in propInfo:
                propInfo["byValueCount"] = Counter()
            propInfo["byValueCount"][dtToCount] += 1
            if self.__createDateProp != prop: # only weekday for this
                return
            try:
                propInfo["byWeekDay"]
            except:
                propInfo["byWeekDay"] = Counter()
            propInfo["byWeekDay"][dayDT.weekday()] += 1     
                        
        typeData["_total"] += 1
        
        if "_firstIEN" not in typeData and len(set(resource).intersection(["_id", "id"])):
            idProp = "id" if "id" in resource else "_id"
            self.__typeData["_firstIEN"] = resource[idProp].split("-")[1]
                    
        # only applies to LIST props - full ones filled in below
        if "fmqlStopped" in resource:
            for propStopped in resource["fmqlStopped"]:
                if propStopped not in typeData:
                    typeData[propStopped] = { "count": 0, "type": "LIST"}
                typeData[propStopped]["count"] += 1
                if "cstopped" not in typeData[propStopped]:
                    typeData[propStopped]["cstopped"] = 1
                else:
                    typeData[propStopped]["cstopped"] += 1
                    
        createDate = getCreateDate(resource, self.__createDateProp)
        dateValues = []
        for prop in [key for key in resource if key not in ["id", "_id", "type", "label", "ien", "fmqlStopped"]]: 

            value = resource[prop] 

            if prop not in typeData:
                typeData[prop] = { "count": 0, "type": typeProp(value) }
            info = typeData[prop]
            info["count"] += 1

            if createDate: # add per property including for create property itself
                if "firstCreateDate" in info:
                    if info["firstCreateDate"] > createDate:
                        info["firstCreateDate"] = createDate
                    if info["lastCreateDate"] < createDate:
                        info["lastCreateDate"] = createDate
                else:
                    info["firstCreateDate"] = createDate
                    info["lastCreateDate"] = createDate
            
            qprop = prop if parentPropPath == "" else parentPropPath + "/" + prop
                
            # If info type not same as value type:
            # - [TODO] neither LITERAL
            # - info type is not LITERAL - skip and record bad literal
            # - info type is LITERAL - reset info with bad literal
            typ = typeProp(value)
            if typ != info["type"]:
                if info["type"] != "LITERAL":
                    if "badLiteralValues" not in info:
                        info["badLiteralValues"] = 0
                    info["badLiteralValues"] += 1
                    continue
                info = { "count": 1, "type": typ, "badLiteralValues": info["count"] }
                typeData[prop] = info
                
            if typeData[prop]["type"] in ["LITERAL", "BOOLEAN", "INTEGER"]:
                countByLiteralValue(qprop, typeData[prop], value)
                continue
                
            if typeData[prop]["type"] == "POINTER":
                if "rangeTypes" not in typeData[prop]:
                    typeData[prop]["rangeTypes"] = [] # Allow variable ptr
                rangeType = value["id"].split("-")[0]
                if rangeType not in typeData[prop]["rangeTypes"]:
                    typeData[prop]["rangeTypes"].append(rangeType)
                ptrType = value["id"].split("-")[0]
                if "label" in value and ptrType not in self.__deidentifyPointerTypes:
                    ptrValue = '{} [{}]'.format(value["label"], value["id"])
                else:
                    ptrValue = value["id"]
                countByPointerValue(qprop, typeData[prop], ptrValue)
                continue
                
            if typeData[prop]["type"] == "LIST":
                countListValueByLength(typeData[prop], value)
                if "reduction" not in typeData[prop]:
                    typeData[prop]["reduction"] = {"_total": 0}
                spropPath = prop if parentPropPath == "" else parentPropPath + "/" + prop
                for sresource in resource[prop]:
                    self.__reduce(sresource, typeData[prop]["reduction"], spropPath)
                continue
                
            if typeData[prop]["type"] == "DATE":
                countDateValue(prop, typeData[prop], value["value"])
                dateValues.append((prop, value["value"]))

        # Post 1: first and last date values seen
        if len(dateValues) > 1: # capture which date property is first/last. Can vary. (effects Orders etc)
            if "_firstDateProps" not in typeData:
                typeData["_firstDateProps"] = {}
                typeData["_lastDateProps"] = {}
            dateValues = sorted(dateValues, key=lambda x: x[1])
            if dateValues[0][0] not in typeData["_firstDateProps"]:
                typeData["_firstDateProps"][dateValues[0][0]] = 0 
            typeData["_firstDateProps"][dateValues[0][0]] += 1
            if dateValues[-1][0] not in typeData["_lastDateProps"]:
                typeData["_lastDateProps"][dateValues[-1][0]] = 0
            typeData["_lastDateProps"][dateValues[-1][0]] += 1
            
# ISO‑8601 form allow fast lexical comparison - datetime comparison is orders
# of magnitude slower. This isn't an exact day validity checker but its good
# enough - split on lex is faster than regexp '(20|19)\d\d-[01]\d-[0-3]\d$
def validDayDT(dayValue, minYear=1980, maxYear=2030):
    try:
        dayDT = datetime.strptime(dayValue, "%Y-%m-%d")
    except:
        return None
    yr = int(dayValue.split("-")[0])
    if yr > maxYear or yr < minYear:
        return None
    return dayDT



#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2017-2020 caregraf

import os
import sys
import json
import shutil
from datetime import datetime
import re
from collections import defaultdict
from statistics import median

from ..cacher.cacherUtils import FMQLReplyStore, convert_size, DATAV1_LOCN_TEMPL, DATA_LOCN_TEMPL

"""
TODO: 
- ZERO files -- tied that in (and compare with reports of 0 but not etc)
  - already reports < 0!
- PARTIAL FILES AS MAX HIT 
  - tie that in
- CUT down and just have cache completelness report
- hone as "mu's" and consider what should be in "reusable" utils
- store time in replies --- use ... median time taken as well as size!
        jreply["queryTook"] = queryTook # time taken (queryCached time is set in store)
        # note time as UTC (can change to local times in a display)
        jreply["queryCached"] = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")


Report state of data in DataV1 | Data Cache to:
- Note degree of completeness: first thing to run ie/ all to do 
- Suggest better LIMIT settings
  - LIMIT to get maximum reply size of ~8.5M
  - median << biggest => cstops in there so very "peaky"
"""
def reportCacheHealth(stationNumber, isV1):

    print("# Reporting Cache Health for {}\n".format(stationNumber))

    reportCompleteness(stationNumber, isV1)   
    reportLimitToSize(stationNumber, isV1)

    print("")
    
"""
Using ReplyStore basic methods, report on its completeness
"""
def reportCompleteness(stationNumber, isV1): # based on FM numbers

    """
    Clinical PATIENT+ (not audits or a/cs etc) - overlap with VPR's source files
    but has extras like 601_85 (Mental Health) and Appointments isn't explicit
    as 2.98 is a multiple of 2
    
    Note: 702 may NOT be one but it is referenced by VPR file VPRDMC.m
    """
    CLINPATIENTTYPES = [
        
        "100", "9000010", "9000010_18", "8925", "409_68", "120_5", "9000010_06", "9000010_07", "9000010_23", "52", "601_85", "27_11", # top 40 (majority not)
        
        "2", "120_8", "9000011", "63", "68", "55", "70", "74", "75_1", "9000010_11", "9000010_12", "704_117", "405", "130", "123", "26_13", "702"
                
    ]

    dm1Location = DATAV1_LOCN_TEMPL.format(stationNumber) if isV1 else DATA_LOCN_TEMPL.format(stationNumber)
    fmqlReplyStore = FMQLReplyStore(dm1Location, makeDir=False)
    availableTypes = fmqlReplyStore.availableTypes()
    stoppedTypes = fmqlReplyStore.stoppedTypes()
    countByType = fmqlReplyStore.expectedTypeCounts()
    
    MAXTHRESHOLDLABEL = "MORE THAN TEN MILLION"
    currentThreshold = MAXTHRESHOLDLABEL
    typByThreshold = defaultdict(list)
    thresholds = [[10000000, "< TEN MILLION", False], [1000000, "< ONE MILLION", False], [100000, "< ONE HUNDRED THOUSAND", False], [50000, "< FIFTY THOUSAND", False], [10000, "< TEN THOUSAND", False], [1000, "< ONE THOUSAND", False], [100, "< ONE HUNDRED", False], [10, "< TEN", False], [2, "SINGLETONS", False], [1, "ZERO", False], [0, "LESS THAN ZERO (WRONG)", False]]
    print("\n## Completeness\n")
    totalRecords = sum(countByType[typ]["count"] for typ in countByType if countByType[typ]["count"] > 0)
    cachedSoFar = 0
    mu = "\# | Type | Expected Size | DONE?\n--- | --- | --- | --- \n"
    for i, typ in enumerate(sorted(countByType, key=lambda x: countByType[x]["count"], reverse=True), 1):
        for thresInfo in thresholds:
            if not thresInfo[2]:
                if countByType[typ]["count"] < thresInfo[0]:
                    mu += "&nbsp; | &nbsp; | __{}__ | &nbsp;\n".format(thresInfo[1])
                    thresInfo[2] = True  
                    currentThreshold = thresInfo[1]
                break
        typByThreshold[currentThreshold].append(typ)
        if typ in availableTypes:
            doneMU = "__YES__" if typ not in stoppedTypes else "__YES (STOPPED)__"
            cachedSoFar += countByType[typ]["count"] # crude as only partial counts as whole
        else:
            doneMU = ""
        labelMU = countByType[typ]["label"] if typ not in CLINPATIENTTYPES else "__{}__".format(countByType[typ]["label"])
        countMU = "{:,}".format(countByType[typ]["count"])
        if countByType[typ]["count"] > 0:
            percCount = float(countByType[typ]["count"])/float(totalRecords) * 100
            if percCount > 0.009:
                countMU += "/{:.2f}%".format(percCount)
        mu += "{} | {} ({}) | {} | {}\n".format(i, labelMU, typ, countMU, doneMU)
    mu += "\n"
    mu += "__Cached so far:__ < {:,} out of {:,} records (fileman count)\n\n".format(cachedSoFar, totalRecords)
    print(mu)
    print("")
    print("Types by threshold (can use in configs)")
    print("")
    thresholds.insert(0, [-1, MAXTHRESHOLDLABEL])
    for i, thresInfo in enumerate(thresholds, 1):
        thresId = thresInfo[1]
        cntRecordsInThreshold = sum(countByType[typ]["count"] for typ in typByThreshold[thresId])
        cntRecordsInThresholdC = cntRecordsInThreshold if cntRecordsInThreshold > 0 else 0
        percRecordsInThreshold = float(cntRecordsInThresholdC)/float(totalRecords) * 100
        print("{}. {} ({}/{:,}/{:.2f}%) - {}".format(i, thresId, len(typByThreshold[thresId]), cntRecordsInThresholdC, percRecordsInThreshold, json.dumps(typByThreshold[thresId])))
        print("")
    print("")
    print("")
    
def reportLimitToSize(stationNumber, isV1):

    print("## Limit to Size Issues ({} Form)".format("V1" if isV1 else "V10"))
    print("")
    
    dm1Location = DATAV1_LOCN_TEMPL.format(stationNumber) if isV1 else DATA_LOCN_TEMPL.format(stationNumber)
    fmqlReplyStore = FMQLReplyStore(dm1Location, makeDir=False)
            
    sizeByFileTypeByFile = fmqlReplyStore.replySizeByType()
                    
    TARGET_MAX_REPLY_IN_BYTES = 8900000 # 8.5MB - raw or actual, not zipped
    
    typsWithNewLimits = {}
    amu = "Type | Limit | \# Replies | Median | Average | Biggest | Fraction | New Limit\n--- | --- | --- | --- | --- | ---\n"
    for flTyp in sorted([flTyp for flTyp in sizeByFileTypeByFile], key=lambda x: float(re.sub(r'\_', '.', x))):
        sys.stdout.write('..{}'.format(flTyp))
        sys.stdout.flush()
        sizes = [sizeByFileTypeByFile[flTyp][fl]["actual"] for fl in sizeByFileTypeByFile[flTyp]]
        mdian = median(sizes)
        biggest = max(sizes)
        avg = sum(sizes)/float(len(sizes))
        noReplies = len(sizeByFileTypeByFile[flTyp])
        fraction = round((float(biggest) / float(TARGET_MAX_REPLY_IN_BYTES)), 1)
        """
        Don't be too exact - within 0.9 to 1.1 of the size is fine
        and if only one reply that's small - then that's fine
        """
        if (noReplies > 1 and (fraction > 1.1 or fraction < 0.9)) or (noReplies == 1 and fraction > 1.1):
            # get limit from last reply
            lastReply = fmqlReplyStore.lastReplyOfType(flTyp)
            limit = int(lastReply["fmql"]["LIMIT"])
            newLimit = int(round(limit / fraction, 0)) 
            mu = "{flTyp} | {limit} | {noReplies} | {mdian} | {avg} | {biggest} | {fraction} | {newLimit}\n".format(flTyp=flTyp, limit=limit, noReplies=noReplies, mdian =convert_size(mdian), avg=convert_size(avg), biggest=convert_size(biggest), fraction=fraction, newLimit=newLimit)
            amu += mu
            typsWithNewLimits[flTyp] = str(newLimit)
            
    print("")
    print("Total Files: {} - One Reply Files: {} - Need new Limit Files: {}".format(len(sizeByFileTypeByFile), sum(1 for flTyp in sizeByFileTypeByFile if len(sizeByFileTypeByFile[flTyp]) == 1), len(typsWithNewLimits)))
    print("Cache uses Zip (size is actual so no effect): {}".format(fmqlReplyStore.usesZip()))
    print("Size threshold: {}".format(convert_size(TARGET_MAX_REPLY_IN_BYTES)))
    print("")
    print(amu)
    print("")
    print("List of types with new limits: {}".format(json.dumps(typsWithNewLimits)))
    print("... add to Cacher config LIMIT setting")
    print("")
    print("")

# ############################# Driver ####################################

def main():

    assert sys.version_info >= (3, 3)

    USAGE = "Usage _EXE_ STATIONNO [V1]"
    try:
        stationNumber = sys.argv[1]
    except IndexError:
        raise SystemExit(USAGE)

    isV1 = False
    if len(sys.argv) == 3:
        if sys.argv[2] == "V1":
            isV1 = True
        else:
            raise SystemExit(USAGE)

    reportCacheHealth(stationNumber, isV1)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2017-2020 caregraf

import os
import json
from datetime import datetime, timedelta
import pytz
import re
from collections import defaultdict
import math
import logging
from logging.handlers import RotatingFileHandler
from logging import handlers
import zipfile

from .. import VISTA_DATA_BASE_DIR
BASE_LOCN_TEMPL = VISTA_DATA_BASE_DIR + "{}/"
DATAV1_LOCN_TEMPL = BASE_LOCN_TEMPL + "DataV1/"
DATA_LOCN_TEMPL = BASE_LOCN_TEMPL + "Data/"
DATARF_LOCN_TEMPL = BASE_LOCN_TEMPL + "DataRF/"
SCHEMA_LOCN_TEMPL = BASE_LOCN_TEMPL + "Schema/"
LOG_LOCN_TEMPL = BASE_LOCN_TEMPL + "Logs/"
TMPWORKING_LOCN_TEMPL = BASE_LOCN_TEMPL + "TmpWorking/"
DATAMETA_LOCN_TEMPL = BASE_LOCN_TEMPL + "DataMeta/"

PATIENT_NAME_TYPES = ["2", "9000001", "55", "63", "70", "631", "665", "690", "790", "38_1"]

"""
Really FMQL[Data]ReplyStore for DESCRIBE replies using afterien, the basic reply of the Cacher. Navigates choice of raw JSON or zip based storage

Implementation note - IEN ordering: 
- don't use split(".") when extracting afterien from file names as some IENs (3_081) are floats. Hence use of re.sub on file suffix.
- open question: any NON FLOAT/NON INTEGER IENs? Enforces none in 'flush'.

TODO: 
- Ability to reset replies for new LIMIT ie/ relimit replies of type
- FMQLSchemaStore to parallel FMQLReplyStore (may become FMQLDataStore)
- MOVE this and other utilities to just take 'stationNumber' ala other cacher utils - ie/ force the /.../Data[V1] structure
"""
class FMQLReplyStore:

    # useZip only applies to EMPTY or newly created
    def __init__(self, cacheLocation, useZip=False, makeDir=True):
    
        """
        Test for both type of cache (ZIP or RAW) and if good or not (BAD_)
        """
        def cacheType(cacheLocation):
            if sum(1 for f in os.listdir(cacheLocation)) == 0:
                return "EMPTY"
            zfs = set(zf for zf in os.listdir(cacheLocation) if re.search(r'\.zip', zf))
            if len(zfs):
                if sum(1 for zf in zfs if not re.match(r'[\d\_]+(\-S)?\-[\d\.]+\.zip$', zf)):
                    return "BAD_ZIP"
                return "ZIP"
            jfs = set(jf for jf in os.listdir(cacheLocation) if re.search(r'\.json', jf))
            if len(jfs):
                if sum(1 for bjf in jfs if not re.match(r'[\d\_]+(\-S)?\-[\d\.]+\.json$', bjf)):
                    return "BAD_RAW"
                return "RAW"
            return "BAD_NOTEMPTY"
    
        if not os.path.isdir(cacheLocation): 
            if not makeDir:
                logging.error("No cacheLocation {}".format(cacheLocation))
                raise Exception("No cacheLocation {}".format(cacheLocation))
            os.mkdir(cacheLocation)
            logging.info("Created location {} for Replies".format(cacheLocation))
            self.__isZip = useZip
        else:
            cType = cacheType(cacheLocation)
            if re.match(r'BAD', cType):
                raise Exception("Bad cache {} - {}".format(cacheLocation, cType))
            if cType == "EMPTY":
                self.__isZip = useZip
            else:
                self.__isZip = True if cType == "ZIP" else False 
        self.__cacheLocation = cacheLocation if re.search(r'\/$', cacheLocation) else cacheLocation + "/"
        
    def usesZip(self):
        return self.__isZip

    def availableTypes(self):
        return set(f.split("-")[0] for f in self.replyFiles())

    def stoppedTypes(self):
        return set(f.split("-")[0] for f in self.replyFiles() if re.search(r'\-S\-', f))
        
    """
    From SELECT TYPES - expected to be in {cacheLocation}/Schema
    
    Use [a] report on FM # vs real and [b] clump types based on size for 
    cacher strategy [c] see progress/ what's left (combine with
    _availableTypes_
    
    Note: any counts < 0 turned into -1
    """
    def expectedTypeCounts(self):
        schemaLocation = re.sub(r'[^\/]+\/$', '/', self.__cacheLocation)
        selectTypes = json.load(open("{}/Schema/SELECT_TYPES.json".format(schemaLocation)))
        expectedCountByType = {}
        for result in selectTypes["results"]:
            if "parent" in result:
                continue
            typ = re.sub(r'\.', '_', result["number"])
            if "count" not in result:
                expectedCountByType[typ] = {"label": result["name"], "count": 0}
                continue
            cnt = int(result["count"]) if not re.search(r'\-', result["count"]) else -1
            expectedCountByType[typ] = {"label": result["name"], "count": cnt}
        return expectedCountByType
        
    def replyFiles(self):
        suffix = "zip" if self.__isZip else "json"
        return [f for f in os.listdir(self.__cacheLocation) if os.path.isfile(os.path.join(self.__cacheLocation, f)) and re.search('\.' + suffix + '$', f)]
        
    """
    Only works if reply files follow IENs names (see below)
        format(reply["fmql"]["TYPE"], reply["fmql"]["AFTERIEN"], "zip") 
        
    Note: coming is related 'replyFileOfCreateDate'
    """
    def replyFileOfResource(self, resourceId):
        replyType = resourceId.split("-")[0]
        sreplyFiles = sorted([f for f in self.replyFiles() if re.match(replyType + "\-", f)], key=lambda x: float(re.sub(r'\.(json|zip)$', '', x.split("-")[-1])))
        if len(sreplyFiles) == 1:
            return sreplyFiles[0] # note: not checking inside so possible NOT!
        resourceIENF = float(resourceId.split("-")[1])
        if resourceIENF == 0:
            return sreplyFiles[0]
        for i, replyFile in sreplyFiles:
            afterIEN = float(re.sub(r'\.(json|zip)$', '', replyFile.split("-")[-1]))
            if resourceIENF <= afterIEN:
                return sreplyFiles[i-1]
        return "" # can't find
        
    """
    Assumption - date order is reply file/resource order. Hence 'create day'

    Use: process only resources made after a certain date.
    
    Note: 'day' is granularity, not time.
    
    Use in recipe for caching more from fresh image of semi-state types
    like 3.081 (just need close) or 2005. Get firstReplyFileOnOrAfterCreateDay
    and take the one before it as source of afterien
    """
    def firstReplyFileOnOrAfterCreateDay(self, replyType, dtProp, dayValue):
        if not isinstance(dayValue, datetime):
            dayValue = datetime.strptime(dayValue, "%Y-%m-%d")
        sreplyFiles = self.replyFilesOfType(replyType)
        if len(sreplyFiles) == 0:
            return ""
        lastReplyFileLoaded = ""
        def getDay(resource, dateProp):
            if dateProp not in resource:
                return ""
            # Allow for real date and a pointer who's label is a date!
            if isinstance(resource[dateProp], dict):
                if "value" in resource[dateProp]:
                    return resource[dateProp]["value"].split("T")[0]
                elif "label" in resource[dateProp]:
                    return resource[dateProp]["label"].split("T")[0]
            return ""
        while True:
            if len(sreplyFiles) == 0: # no reply at or above that date
                firstReplyFile = ""
                break
            midIndex = int(len(sreplyFiles)/2)
            if sreplyFiles[midIndex] == lastReplyFileLoaded:
                if midIndex == 0:
                    raise Exception("Stuck on 0 - should never happen")
                midIndex = midIndex - 1 # move down one (see below)
            lastReplyFileLoaded = sreplyFiles[midIndex]
            midReply = self.loadReplyFromFile(sreplyFiles[midIndex])
            lastDayValue = None
            for resource in reversed(midReply["results"]):
                lastDayValueStr = getDay(resource, dtProp)
                if lastDayValueStr == "":
                    continue
                try:
                    lastDayValue = datetime.strptime(lastDayValueStr, "%Y-%m-%d")
                except:
                    logging.info("Skipping resource of reply {} as bad date value {}".format(sreplyFiles[midIndex], lastDayValueStr))
                    continue
                break
            # For now insist all replies have at least one resource with prop
            if lastDayValue == None:
                # sreplyFiles.pop(midIndex)
                # continue
                firstReplyFile = ""
                break
            if lastDayValue < dayValue: # delete all before and inc this one
                sreplyFiles = sreplyFiles[midIndex+1:]
                continue
            # >= and first one! => this is it or as good as we get
            if midIndex == 0:
                firstReplyFile = sreplyFiles[midIndex]
                break
            # delete all ABOVE this one. If none above, the midIndex
            # check above that avoids repetition will catch it
            sreplyFiles = sreplyFiles[0:midIndex+1]
        return firstReplyFile           
                
    """
    Record both actual and compressed size if zip cache (zip supports it)
    """
    def replySizeByType(self):        
        sizeByFileTypeByFile = defaultdict(lambda: defaultdict(int))    
        for fl in self.replyFiles():
            flTyp = fl.split("-")[0]
            if self.usesZip():
                jfl = re.sub(r'zip$', 'json', fl)
                zfName = "{}/{}".format(self.__cacheLocation, fl)
                zf = zipfile.ZipFile(zfName, "r")   
                jflZInfo = zf.getinfo(jfl) 
                sizeByFileTypeByFile[flTyp][fl] = {"actual": jflZInfo.file_size, "compressed": jflZInfo.compress_size}
            else:
                sizeByFileTypeByFile[flTyp][fl] = {"actual": os.path.getsize(self.__cacheLocation + fl)}
        return sizeByFileTypeByFile

    """
    For cacher - allows it to hone in on LIMIT and AFTERIEN of last reply so can 
    issue next one
    """
    def lastReplyOfType(self, replyType): 
        filesSoFar = [f for f in self.replyFiles() if re.match(replyType + "\-", f)]   
        if len(filesSoFar):
            fl = sorted(filesSoFar, key=lambda x: float(re.sub(r'\.(json|zip)$', '', x.split("-")[-1])))[-1]
            if re.search(r'\.zip$', fl):
                jfl = re.sub(r'zip$', 'json', fl)
                zfName = "{}/{}".format(self.__cacheLocation, fl)
                zf = zipfile.ZipFile(zfName, "r")   
                lastReply = json.loads(zf.read(jfl))
                return lastReply
            else:
                return json.load(open(self.__cacheLocation + fl))
        return None 

    def replyFilesOfType(self, replyType):
        filesSoFar = sorted([f for f in self.replyFiles() if re.match(replyType + "\-", f)], key=lambda x: float(re.sub(r'\.(json|zip)$', '', x.split("-")[-1])))
        return filesSoFar
        
    def loadReplyFromFile(self, fl):
        if re.search(r'\.zip$', fl):
            zf = zipfile.ZipFile(self.__cacheLocation + fl, "r")
            jfl = re.sub(r'zip$', 'json', fl)
            flJSON = json.loads(zf.read(jfl))
        else:
            flJSON = json.load(open(self.__cacheLocation + "/" + fl))
        return flJSON
        
    def totalReplies(self):
        return len(self.replyFiles())
     
    """           
    Generally TYP-AFTERIEN. but if stopped (v1.3 on) then TYP-S-AFTERIEN.
    """
    def flush(self, reply):     
        # Expect FLOAT or INTEGER IEN so that sorting will work
        if not re.match(r'[\d\.]+$', str(reply["fmql"]["AFTERIEN"])):
            raise Exception("Ordering and processing of Replies in Cache rely on integer or float IENs - not {}".format(reply["fmql"]["AFTERIEN"]))
        if len(reply["results"]) == 0:
            raise Exception("Not allowed flush an empty reply - certain cache routines expect at least one result in each cached reply")
        stoppedDesignation = "-S" if ("stopped" in reply or "fmqlMaxStopped" in reply) else ""
        if self.__isZip:
            fl = "{}{}-{}.{}".format(reply["fmql"]["TYPE"], stoppedDesignation, reply["fmql"]["AFTERIEN"], "zip")    
            zfName = "{}/{}".format(self.__cacheLocation, fl)
            zf = zipfile.ZipFile(zfName, "w", zipfile.ZIP_DEFLATED, allowZip64=True) 
            jfl = re.sub(r'zip$', 'json', fl)
            zf.writestr(jfl, json.dumps(reply))       
        else:
            fl = "{}{}-{}.{}".format(reply["fmql"]["TYPE"], stoppedDesignation, reply["fmql"]["AFTERIEN"], "json")
            json.dump(reply, open(self.__cacheLocation + fl, "w"), indent=4)
        return fl # return the file name
        
    """
    Used to avoid recaching - note: .json or .zip independent. ie/ one or other there
    then that suffices.
    """
    def replyFileExists(self, replyFile): 
        suffix = "zip" if self.__isZip else "json"  
        fl = re.sub(r'(json|zip)$', suffix, replyFile)
        if fl in set(self.replyFiles()):
            return fl
        return ""
        
    """
    Records if another process is using/writing to a Reply file
    
    TODO: for parallel operations in multi-threaded operations
    """
    def replyFileInUse(self, replyFile):
        pass
        
    def iterator(self, onlyTypes=None, excludeRepliesFilter=None, startAtReply=""):
    
        class FMQLReplyIterator:
        
            def __init__(self, cacheLocation, onlyTypes, excludeRepliesFilter, startAtReply, isZip):
                
                self.__cacheLocation = cacheLocation
                self.__excludeRepliesFilter = excludeRepliesFilter 
                self.__startAtReply = startAtReply
                
                self.__currentReplyFile = ""
                
                suffix = "zip" if isZip else "json"     
                self.__flsByTyp = defaultdict(list)
                for fl in os.listdir(cacheLocation):
                    if not re.search(suffix + '$', fl):
                        continue
                    flTyp = fl.split("-")[0]
                    # distinguish [] and None
                    if onlyTypes != None and flTyp not in onlyTypes:
                        continue
                    self.__flsByTyp[flTyp].append(fl)
                    
                if len(self.__flsByTyp) == 0:
                    if len(onlyTypes):
                        raise Exception("Can't build iterator when no Replies of Type(s) {} exist".format(", ".join(list(onlyTypes))))
                    raise Exception("Can't build iterator when no Replies exist")
                    
            def currentReplyFile(self):
                return self.__currentReplyFile
                        
            """
            TODO: full iterator
            - startAtReply could be passed to iter optionally 
            - next takes over what iter does but must raise StopIteration() 
            """                
            def __iter__(self):
        
                for typ in sorted(self.__flsByTyp, key=lambda x: float(re.sub("_", ".", x))):

                    for fl in sorted(self.__flsByTyp[typ], key=lambda x: float(re.sub(r'\.(json|zip)$', '', x.split("-")[-1]))):
                                        
                        # Usually only used with one type being iterated
                        if self.__startAtReply != "":
                            if self.__startAtReply == fl:
                                self.__startAtReply = ""
                            else:
                                continue 

                        if self.__excludeRepliesFilter and self.__excludeRepliesFilter(fl):
                            continue # put here so in order
                            
                        self.__currentReplyFile = fl
                        
                        if re.search(r'\.zip$', fl):
                            try:
                                zf = zipfile.ZipFile(self.__cacheLocation + fl, "r")
                            except:
                                raise Exception("Can't handle bad reply ZIP file {}".format(fl))
                            jfl = re.sub(r'zip$', 'json', fl)
                            flJSON = json.loads(zf.read(jfl))
                        else:
                            flJSON = json.load(open(self.__cacheLocation + "/" + fl))            
                    
                        yield flJSON
              
        if startAtReply != "" and not (onlyTypes and len(onlyTypes) == 1):
            raise Exception("Cannot have iterate with 'startAtReply' if don't specify one and only one type")      
        return FMQLReplyIterator(self.__cacheLocation, onlyTypes, excludeRepliesFilter, startAtReply, self.__isZip)
        
"""
Simple resource iterator built over the FMQLReplyStore's iterator

Note on memory use:

    import psutil
    process = psutil.Process(os.getpid())
    print("Memory Use End", process.memory_info().rss, process.memory_percent())
    
can see 2.7% -> 3.5% just doing a no op iteration of a large file. GC has its own mind.
Using ORDERED is both slower and more memory intensive (6.5%).
"""
class FilteredResultIterator:
    
    def __init__(self, replyLocation, typId, filt=None, startAtReply=""):
    
        fmqlReplyStore = FMQLReplyStore(replyLocation)        
        self.__replyIter = fmqlReplyStore.iterator(onlyTypes=[typId], startAtReply=startAtReply)
        self.__filt = filt
            
        self._numberFiltered = 0
            
    @property
    def numberFiltered(self):
        return self._numberFiltered
            
    def __iter__(self):
        for reply in self.__replyIter:
            for resource in reply["results"]:
                if self.__filt and not self.__filt(resource):
                    self._numberFiltered += 1
                    continue
                yield resource
                
    def currentReplyFile(self):
        return self.__replyIter.currentReplyFile()

"""
File by file, cache data and will restart. Checks what's already in store.

TODO: may move into an FMQLCacher that provides caching utilities (as opposed
to the reading of FMQL[Data|Reply]Store.
    
Mandatory arguments:
- fmqlIF
- fmqlReplyStore
- fileType: ex/ 2, 120_5

Optional/defaulted arguments:
- limit for query: defaults to 1000
- cstop for query: defaults to 10
- filter for query: default is none
- maxNumber: maximum number to retrieve. Default is no limit (-1)
- afterIEN (for restart if necessary and for doing LIMIT at a time)
- epWord: query used in CSP and node; Apache uses "fmql"

To make FMQLIF:
    
    from ..fmqlIF.fmqlIF import FMQLRESTIF, FMQLBrokerIF
    
        fmqlIF = FMQLRESTIF(fmqlEP, epWord="fmqlQuery")
                            or
        fmqlIF = FMQLBrokerIF(hostname, port, access, verify, osehraVISTA=False)

Returns:
    COMPLETE
    HITMAX
    NO_SUCH_FILE
    EMPTY_FILE
"""                
def cacheFileToStore(fmqlIF, fmqlReplyStore, fileType, limit=500, cstop=1000, maxNumber=-1):

    queryTempl = "DESCRIBE " + fileType + " LIMIT %(limit)s AFTERIEN %(afterien)s CSTOP " + str(cstop)
    
    # Three cases for AFTERIEN:
    # - no reply => afterien is 0
    # - last reply is the last (# results < limit of its query)
    # - last reply's last result supplies afterien
    # ... note that Cache depends on float or integer IENs and enforces that form
    lastReply = fmqlReplyStore.lastReplyOfType(fileType)
    if not lastReply:
        afterien = 0
        numberOfTypeCachedOverall = 0
    else:
        # REM: limit for next go may not be the same so not overriding limit
        lastLimit = int(lastReply["fmql"]["LIMIT"])
        if lastLimit > len(lastReply["results"]):
            logging.info("Got all of {} already - moving on".format(fileType))
            return "COMPLETE"
        if maxNumber != -1:
            repliesSoFar = fmqlReplyStore.replyFilesOfType(fileType)
            cachedSoFar = len(repliesSoFar) * lastLimit
            if cachedSoFar >= maxNumber:
                logging.info("Cached so far of {} already exceeds max number of {} - moving on".format(cachedSoFar, maxNumber))
                return "HITMAX"
        lastResult = lastReply["results"][-1]
        afterien = (lastResult["_id"] if "_id" in lastResult else lastResult["uri"]["value"]).split("-")[1]
        numberOfTypeCachedOverall = lastLimit * len(fmqlReplyStore.replyFilesOfType(fileType))
        logging.info("Filetype {}, Cached {:,} already, some left to get. Restarting with AFTERIEN {}".format(fileType, numberOfTypeCachedOverall, afterien))

    # queryNo and afterIEN are usually 0 but can start again    
    # Loop until there is no more or we reach the maximum
    numberOfTypeCached = 0
    start = datetime.now()
    stopped = False
    queryNo = 0
    while True:

        queryNo += 1
        query = queryTempl % {"limit": limit, "afterien": afterien}
        noCachedMU = "{:,}".format(numberOfTypeCached) if numberOfTypeCachedOverall == numberOfTypeCached else "{:,}/{:,}".format(numberOfTypeCached, numberOfTypeCachedOverall)
        logging.info("Sending query number {} after {} cached, taking {} so far - {}".format(queryNo, noCachedMU, datetime.now() - start, query))

        queryStart = datetime.now()
        jreply = fmqlIF.invokeQuery(query)
        if "error" in jreply:
            # Shouldn't happen for Cacher as it walks Schema but for some VistAs (FOIA)
            # definitions may be there though files are not?
            if jreply["error"] == "No such file": # will treat like a "Zero file"
                if queryNo != 1:
                    raise Exception("Expect 'no such file' error with first query")
                logging.info("Bailing on {} query as no such file".format(fileType))
                return "NO_SUCH_FILE"
            logging.error("Received error reply: {}".format(jreply["error"]))
            raise Exception(jreply["error"])

        # Special case - first call (afterien=0) and no results => don't cache
        # ... REM: doing TOPONLY in case POPONLY is wrong. Means more queries but safer.
        # Ex for CHCS Synth: ['3_081', '66', '52'] not in POP but have "data"
        # ... REM: afterien != 0 and no results then still cache as it means the case of
        # second to last LIMITED query filled up and then the last just returns none.
        # Note: alt is do a COUNT and then DESCRIBE if need be but largely the same cost.
        if afterien == 0 and len(jreply["results"]) == 0:
            logging.info("Empty {} - 0 replies afterien 0".format(fileType))
            return "EMPTY_FILE" 

        # Don't return False (<=> empty file) - but no results left ... happens in rare 
        # case of LIMIT coinciding with end of previous reply
        if len(jreply["results"]) == 0:
            break

        queryTook = str(datetime.now() - queryStart)
        jreply["queryTook"] = queryTook # time taken (queryCached time is set in store)
        # note time as UTC (can change to local times in a display)
        jreply["queryCached"] = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
        if "fmqlMaxStopped" in jreply or "stopped" in jreply:
            stopped = True
        try:
            fmqlReplyStore.flush(jreply)
        except:
            logging.error("Failed to serialize {}/{} - backing out".format(fileType, afterien))
            raise
        logging.info("Queried and flushed {} afterien {} in {} with {} resources, any stopped {}".format(fileType, str(afterien), queryTook, len(jreply["results"]), "**TRUE**" if stopped else "FALSE"))
        
        numberOfTypeCached += len(jreply["results"])
        numberOfTypeCachedOverall += len(jreply["results"])
        if len(jreply["results"]) != int(limit):
            break

        # TODO: properly reset limit at the start to make sure maximum never exceeded
        if maxNumber != -1 and numberOfTypeCached >= maxNumber:
            logging.debug("Breaking as got or exceeded maximum requested - {} - for {}".format(maxNumber, fileType))
            break
        lastResult = jreply["results"][-1]
        afterien = (lastResult["_id"] if "_id" in lastResult else lastResult["uri"]["value"]).split("-")[1]
        if (queryNo % 100) == 0:
            logging.debug("So far this has taken {}".format(datetime.now() - start))

    noCachedMU = "{:,}".format(numberOfTypeCached) if numberOfTypeCachedOverall == numberOfTypeCached else "{:,}/{:,}".format(numberOfTypeCached, numberOfTypeCachedOverall)
    logging.info("Finished - cached {}, stopped {}, took {}".format(noCachedMU, "TRUE" if stopped else "FALSE", datetime.now() - start))
    
    return "COMPLETE"
   
"""
Inside [cacheLocation], will zip all replies of type [fileType] AND remove the unzipped replies. Allows retroactive zipping
... could be simpler - type then reply file just to do the work in order
""" 
def zipAllReplies(cacheLocation):

    def zipRepliesOfType(fileType, cacheLocation):
        start = datetime.now()
        if cacheLocation[-1] != "/":
            cacheLocation += "/"
        filesOfType = [f for f in os.listdir(cacheLocation) if os.path.isfile(os.path.join(cacheLocation, f)) and re.search('\.json$', f) and re.match(fileType + "\-", f)]
        if len(filesOfType) == 0:
            logging.info("No files of type {} to archive - exiting".format(fileType))
            return
        for i, f in enumerate(filesOfType, 1):
            qf = os.path.join(cacheLocation, f)
            with zipfile.ZipFile(cacheLocation + re.sub(r'json$', 'zip', f), "w", zipfile.ZIP_DEFLATED, allowZip64=True) as typeZip:
                typeZip.write(qf, f)
                os.remove(qf)
        logging.info("Wrote {} files of type {} to archives in {}".format(i, fileType, str(datetime.now() - start)))
        
    start = datetime.now()
    
    flTyps = set()
    for fl in os.listdir(cacheLocation):
        if not re.search(r'\.json$', fl):
            continue
        flTyp = fl.split("-")[0]
        flTyps.add(flTyp)
    for i, flTyp in enumerate(flTyps, 1):
        zipRepliesOfType(flTyp, cacheLocation)
        
    logging.info("Wrote {} types to archives in {}".format(i, str(datetime.now() - start)))
    
"""
Or
   for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0
    return num
"""
def convert_size(size_bytes):
   if size_bytes == 0:
       return "0B"
   size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
   i = int(math.floor(math.log(size_bytes, 1024)))
   p = math.pow(1024, i)
   s = round(size_bytes / p, 2)
   return "%s %s" % (s, size_name[i])
   
"""
Deprecated: older V2 from V1 FMQL. V3 FMQL now largely embeds the form created here. Here for old dataset transformation
leveraging fmqlV2er
"""
class V1RecordToV2:

    # actualTimeZone from "for tz in pytz.all_timezones" ex/ US/Mountain
    def __init__(self, forMongo=False, actualTimeZone=""):

        ID = "V1RecordToV2"
        DESCRIPTION = "From DM V1 to DM V2 - DM in 1 was client-side"
        
        self.__idProp = "id" if not forMongo else "_id"
        self.__transformedByTypeCount = defaultdict(int)
        
    def transform(self, record):
        nrecord = self.__reframe(record)
        self.__transformedByTypeCount[nrecord[self.__idProp].split("-")[0]] += 1
        return nrecord
        
    def __newType(self, record):
    
        typV2 = record["uri"]["label"].split("/")[0] + "-" + record["uri"]["value"].split("-")[0]

        TYP_MAP = {
            "STATION NUMBER TIME SENSITIVE-389_9": "STATION NUMBER (TIME SENSITIVE)-389_9",
            "ICD OPERATION_PROCEDURE-80_1": "ICD OPERATION/PROCEDURE-80_1",
            "USR AUTHORIZATION_SUBSCRIPTION-8930_1": "USR AUTHORIZATION/SUBSCRIPTION-8930_1",
            "OE_RR REPORT-101_24": "OE/RR REPORT-101_24",
            "OE_RR EPCS PARAMETERS-100_7": "OE/RR EPCS PARAMETERS-100_7",
            "NDC_UPN-50_67": "NDC/UPN-50_67",
            "SERVICE_SECTION-49": "SERVICE/SECTION-49",
            "SIGN_SYMPTOMS-120_83": "SIGN/SYMPTOMS-120_83"
        }
        if typV2 in TYP_MAP:
            typV2 = TYP_MAP[typV2]
            
        return typV2 
        
    def __newProp(self, prop, propValue):
        RESERVED_PROPS = ["id", "_id", "type", "value", "parent", "label", "ien"]
        if prop in RESERVED_PROPS:
            return "{}_{}".format(prop, re.sub(r'\.', '_', propValue["fmId"]))
        # Issue: NCName ::= (Letter | '_') (NCNameChar)* in http://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-NCName BUT JSON allows a number etc first. We insert a _ if there is a number first. TBD: best done in FMQL to be consistent.
        if re.match(r'\d', prop):
            return "_" + prop
        return prop
                
    def __reframe(self, record, isMultiple=False):

        if isMultiple == False:
            nrecord = {
                self.__idProp: record["uri"]["value"],
                "type": self.__newType(record)
            }
            self.__topRecord = nrecord # for ref back up of cnodes
        else:
            ien = record["uri"]["value"].split("-")[1].split("_")[0]
            nrecord = {"ien", ien}
            
        """
        Issue: until uses VistA time zone settings, FMQL "Z" (nominally UTC time)
        is actually local time, stampled as UTC. UTC is ideal - GUIs can go local
        but DB should be UTC. But must be real UTC
        
        Using: http://pytz.sourceforge.net/
    
        Ex/ "US/Mountain" ... pytz can list all
        
        TODO:
        - more on daylight savings: is_dst (see docs ... seems to be ambiguity in October)
        - take all logic from toMongoDate ie/ the 24:00:00 to 00:00:00 plus day
        - apply to V2 (over V1) based on passed in time zone
        """
        def fixBadUTCTime(badDTValue, realTZName):
            dt = datetime.strptime(badDTValue, "%Y-%m-%dT%H:%M:%SZ") # Z ignored
            tzinfo=pytz.timezone(realTZName)
            dttzed = tzinfo.localize(dt)
            dtutc = dttzed.astimezone(pytz.utc)
            return dtutc.strftime("%Y-%m-%dT%H-%M-%SZ")

        for prop in record:

            if prop in ["uri"]:
                continue

            propValue = record[prop]            
            nprop = self.__newProp(prop, propValue)

            if propValue["type"] == "cnodes": # fmType == "9"
                # Record stops inline but not they exist in top level (and reply)
                if "stopped" in propValue:
                    # FMQL 1.1 didn't record fmCount so allow for that.
                    try:
                        fmCount = int(propValue["fmCount"])
                    except:
                        fmCount = -1
                    # Note: missing subTypeId in V3 as not available
                    if "fmqlStopped" not in nrecord:
                        nrecord["fmqlStopped"] = {}
                        if "fmqlMaxStopped" not in self.__topRecord:
                            self.__topRecord["fmqlMaxStopped"] = -1
                    # v1.3 added fmCount but can be BAD ie/ <= REPLY CSTOP
                    nrecord["fmqlStopped"][nprop] = {"count": fmCount}
                    # Ensure top record has maximum list size
                    if self.__topRecord["fmqlMaxStopped"] < fmCount:
                        self.__topRecord["fmqlMaxStopped"] = fmCount
                    logging.info("Stopped property {} of {}, count {}".format(prop, record["uri"]["value"], fmCount))
                    continue
                nrecord[nprop] = []
                for i, srecord in enumerate(propValue["value"], 1):
                    nrecord[nprop].append(self.__reframe(srecord, True))
                continue
                        
            # Note: no time zone proper yet -- leaving for V2 FMQL
            if propValue["fmType"] == "1": # datetime
                if re.search(r'T00:00:00Z$', propValue["value"]):
                    nvalue = {"value": propValue["value"].split("T")[0], "type": "xsd:date"}
                else:
                    nvalue = {"value": fix24DateTime(propValue["value"]), "type": "xsd:dateTime"}
            elif propValue["fmType"] == "2": # treat numeric as literal just in case
                nvalue = propValue["value"]
            # or could just 'ivalue' in record[prop]
            elif propValue["fmType"] == "3": # type is literal
                nvalue = "{}:{}".format(propValue["ivalue"], propValue["value"])
            elif propValue["fmType"] == "4": # literal
                nvalue = propValue["value"]
            # 'datatype': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral too
            elif propValue["fmType"] == "5": 
                nvalue = re.sub(r'\r', '\n', propValue["value"])
            # 6: COMPUTED
            elif propValue["fmType"] in ["7", "8"]:
                nvalue = {"id": propValue["value"], "label": propValue["label"].split("/")[1]}
            # 9: MULTIPLE (above)
            elif propValue["fmType"] == "12": # Boolean, datetype == xsd:boolean
                # Fall back to literal (may revisit and suppress)
                if propValue["value"] not in ["true", "false"]:
                    nvalue = propValue["value"]
                else:
                    nvalue = True if propValue["value"] == "true" else False
            # Q C***, K VISTA
            # "10": "MUMPS",
            # "11": "IEN", # IEN match in .001
            else:
                raise Exception("UNEXPECTED TYPE {}".format(propValue["fmType"]))
                
            # Top level record has label <=> first prop (may drop)
            if len(nrecord) == 2 and "type" in nrecord:
                # 8989_5 and  8930_3 have ptr as .01
                if isinstance(nvalue, dict): # pter or date
                    nrecord["label"] = nvalue["label"] if "label" in nvalue else nvalue["value"]
                else: 
                    nrecord["label"] = nvalue
                
            nrecord[nprop] = nvalue
                            
        return nrecord
        
    def getTransformedByTypeCount(self):
        return self.__transformedByTypeCount
                
"""
Recache any stopped resources - exceptions if can't - and restore in a rebuilt
reply.

This ONLY works if DESCRIBING one at a time (top level) can safely get a broad node. From v3, see 'fmqlContainsTotal' (total number of contained elements seen apart from stopping) as well as 'fmqlMaxStopped' at the global level of a Reply
"""
def recacheForCSTOPs(stationNo, typId, fmqlIF, flushToRF=False):

    start = datetime.now()
    logLocn = re.sub(r'Data\/', 'Logs/', DATA_LOCN_TEMPL).format(stationNo)
    configLogging(stationNo, "REMOVECSTOP")
    logging.info("Begin removal of cstops from {}".format(typId))
    dataLocn = DATA_LOCN_TEMPL.format(stationNo)
    dataReplyStore = FMQLReplyStore(dataLocn)  
    # TMP: just to be careful 
    destDataLocn = DATARF_LOCN_TEMPL.format(stationNo) if flushToRF else dataLocn
    rfReplyStore = FMQLReplyStore(destDataLocn, useZip=True)     
    replyIter = dataReplyStore.iterator(onlyTypes=[typId])
    noRecachedStoppedResources = 0
    noRepliesWithStoppedResources = 0
    for i, reply in enumerate(replyIter, 1):
        replyFile = replyIter.currentReplyFile()
        if not re.search(r'\-S', replyFile):
            logging.info("Leaving {} as had NO stops".format(replyFile))
            continue # relies on Marker
        replyHasStoppedResources = False
        nresults = []
        for resource in reply["results"]:
            if "fmqlStopped" in resource:
                replyHasStoppedResources = True
                logging.info("Encountered stopped resource {} in {} - recaching individually".format(resource["_id"], replyFile))
                query = "DESCRIBE {} CSTOP {}".format(resource["_id"], 1000000)
                queryStart = datetime.now()
                resourceReply = fmqlIF.invokeQuery(query)
                logging.info("Re-cache took {}".format(str(datetime.now() - queryStart)))
                if "fmqlMaxStopped" in resourceReply: 
                    raise Exception("Re-query Stopped too {} - exiting".format(resource["_id"])) 
                nresource = resourceReply["results"][0]
                if "fmqlStopped" in nresource: # "legacy" check
                    raise Exception("Internal V3 Error: shouldn't happen as fmqlMaxStopped should stop it")
                noRecachedStoppedResources += 1
                nresults.append(nresource)
            else:
                nresults.append(resource)
        if not replyHasStoppedResources:
            raise Exception("V3 Internal Error: reply file marked with an -S but no stopped resources")
        reply["results"] = nresults
        del reply["fmqlMaxStopped"] # no need any more
        noRepliesWithStoppedResources += 1
        if not flushToRF: # only done if not flushing to RF
            logging.info("Deleting obsolete stopped reply file {}".format(replyFile))
            os.remove("{}{}".format(dataLocn, replyFile))
        logging.info("Reflushing {} w/o stops".format(re.sub(r'\-S', '', replyFile)))
        rfReplyStore.flush(reply)
    finish = datetime.now()
    if noRepliesWithStoppedResources:
        logging.info("FINISHED at {} in {}: recached {:,} resources from {:,} replies after seeing {:,} replies".format(
            finish,
            finish - start,
            noRecachedStoppedResources,
            noRepliesWithStoppedResources,
            i
        ))
    else:  
        logging.info("FINISHED at {} in {} after seeing NO STOPPED RESOURCEs in {:,} replies".format(
            finish,
            finish - start,
            i
        ))
        
"""
metaOfVistA for cutTime and key KSP ids including name - requires 8989_3 and 
PostMaster user needs to be cached first.

Naming:
- name | KSP domain name (first part of URL)
- longName | KSP default institution name
and notes (as expected), default institution IEN matches station number 

Note: older version used last 3.081, 409.84 too and they could be added
optionally later and from Data location.

TODO:
- KSP/production (PROD^XUPROD - prod info in KSP)
  - https://github.com/caregraf/fmqlutils/issues/4 ... about Entity extractors built in now to run to check my results with those
- Last 3.081,
- Build etc DESCRIBE LAST
- Last Exp DESCRIBE LAST 757_01 
- Parameters: any global settings worth (can be) taken? 
  ... S SYS=$S($D(VPRSYS):VPRSYS,1:$$GET^XPAR("SYS","VPR SYSTEM NAME"))
and
> Every Fileman installation has a SITE name and number. These values are
established when initializing Fileman, and are stored in the ^DD("SITE") and
^DD("SITE",1) nodes.
... need to go back to a DESCRIBE SYSTEM?
"""
def metaOfVistA(stationNo):

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "metaOfVistA.json"
    try:
        metaInfo = json.load(open(jsnFl))
    except:
        pass
    else: 
        return metaInfo 

    fmqlReplyStore = FMQLReplyStore(DATAMETA_LOCN_TEMPL.format(stationNo))
    kspReply = fmqlReplyStore.lastReplyOfType("8989_3")
    postMasterReply = fmqlReplyStore.lastReplyOfType("200")
    installReply = fmqlReplyStore.lastReplyOfType("9_7")
    if not (kspReply and postMasterReply and installReply):
        raise Exception("Haven't run fmqlcacher ... META so can't see metaInfo")
    kspResource = kspReply["results"][0]
    postMasterResource = postMasterReply["results"][0]
    firstInstallResource = installReply["results"][0]

    metaInfo = {"stationNo": stationNo}    
    metaInfo["longName"] = kspResource["default_institution"]["label"]
    # ex/ PALO-ALTO.MED.VA.GOV
    metaInfo["domainName"] = kspResource["domain_name"]["label"]
    dnlMatch = re.match(r'([^\.]+)\.MED\.VA\.GOV$', kspResource["domain_name"]["label"])
    if dnlMatch:
        metaInfo["name"] = re.sub(r'-', ' ', dnlMatch.group(1)).title()
    defaultInstitutionIEN = kspResource["default_institution"]["id"].split("-")[1]
    if defaultInstitutionIEN == stationNo:
        metaInfo["isSNODefaultInstitutionIEN"] = True # ie/ 668 Spokane etc and name Spokane taken for long_name
    if "dns_ip" in kspResource:
        metaInfo["dnsIP"] = kspResource["dns_ip"]
    if "production" in kspResource:
        metaInfo["isProduction"] = kspResource["production"]
    for pprop in [("facility_iso", "iso"), ("facility_cio", "cio")]:
        if pprop[0] in kspResource:
            metaInfo[pprop[1]] = kspResource[pprop[0]]

    if "date_entered" in postMasterResource:
        metaInfo["dateEnteredPostMaster"] = postMasterResource["date_entered"]["value"]
    metaInfo["lastSignonPostMaster"]  = postMasterResource["last_signon_date_time"]["value"]
    metaInfo["cutDate"] = metaInfo["lastSignonPostMaster"].split("T")[0]
    
    # probably KERNEL - VIRGIN INSTALL 8.0 ... let's see (and may be very common across
    # VistA's.
    metaInfo["nameFirstInstall"] = firstInstallResource["name"] 
    if "date_loaded" in firstInstallResource:
        metaInfo["dateLoadedFirstInstall"] = firstInstallResource["date_loaded"]["value"]
        
    json.dump(metaInfo, open(jsnFl, "w"), indent=4)
                
    return metaInfo
        
"""
Accounts for 24:00:00 - moves to next day

TMP: will move to V2er
"""
def fix24DateTime(dtStr):
    if not re.search(r'T24:00:00Z?$', dtStr):
        return dtStr
    try:
        dt = datetime.strptime(dtStr.split("T")[0], "%Y-%m-%d")
    except: # don't fix - may be issue as leads to problem above or bad time TODO
        return dtStr
    dt = dt + timedelta(days=1)
    dtStr = datetime.strftime(dt, "%Y-%m-%dT%H:%M:%SZ")
    return dtStr
    
"""
FM's trailing number implies 0 for time so .1 = 10 o clock or
.101 == 10:10 and NOT 01 or 10.01. It also allows 24 ie/ 24:00:00.
For Python etc use, this should become 00:00:00 of the next day

(Off: S Y=3180304.1 X ^DD("DD") W Y)

Tests:
- 3180323 -> 2018-03-23
- 3180323.2 -> 2018-03-23T20:00:00
- 3180323.20 -> 2018-03-23T20:00:00 (FM doesn't have this)
- 3180323.203 -> 2018-03-23T20:30:00 
- 3180323.20304 -> 2018-03-23T20:30:40
- 3180323.02030499 -> 2018-03-23T02:03:04 (shortens)
- 3180323.235959 -> 2018-03-23T23:59:59 
- 3180323.240000 -> 2018-03-24T00:00:00 (start next day)
"""    
def mapFMDateTime(fmDateTime):
    dt = re.sub(r'^3', '20', fmDateTime.split(".")[0])
    if not re.search(r'\.', fmDateTime):
        dt = datetime.strptime(dt, "%Y%m%d")
        return datetime.strftime(dt, "%Y-%m-%d")
    tm = fmDateTime.split(".")[1][0:6]
    if tm == "":
        tm = "000000"
    if len(tm) == 1:
        tm = "{}0".format(tm)
    elif len(tm) == 3:
        tm = "{}{}0".format(tm[0:2], tm[2])
    elif len(tm) == 5:
        tm = "{}{}0".format(tm[0:4], tm[4])
    tm = tm.ljust(6, "0")
    if tm == "240000":
        dt = datetime.strptime(dt, "%Y%m%d")
        dt = dt + timedelta(days=1)
        return datetime.strftime(dt, "%Y-%m-%dT%H:%M:%S")
    return datetime.strptime("{}.{}".format(dt, tm), "%Y%m%d.%H%M%S").strftime("%Y-%m-%dT%H:%M:%S") 
    
"""
EXPERIMENTAL (V2 Form now - will move to FMQL V2's base form/stop support)

Simple V2-based flipping Alg:
- RF == Reframe, DataRF 
- will Exception if CSTOP in ANY reply being walked - expect to be on full replies
- splits off [1] mresources and prop from parent reply (ie/ reflushes from
  original V2) and [2] makes a new mresources reply 
- mresources ID'ed in order seen in resource walk which won't be create
  order. 
- explicitInheritProps so container properties flow down. Example is appt create
date of 44.001 to 44.003; also inherits any container pter/parent pter properties
to cover multiples in multiples.
  
Invoke:
    fmqlIF = FMQLBrokerIF(config["hostname"], config["port"], config["access"], 
config["verify"])
    flip(stationNo, "38_1", "date_time_record_accessed", "38_11", "Date_Time_Record_Accessed", fmqlIF, 

TODO:
  * [ ] multipleProp as container name explicitly (issue of keeping hierarchy inherited?)
  * [ ] want 'multiple label prop' ie/ put label into multiple based on its .01 so just like others and then order of load in python etc won't matter
  * [ ] parent rezip too (right now too many) in RF
  * [ ] use prop name like 'patient' etc for parent (allow pass in)
  * [ ] ease of analysis: add in "count children" property ie/ {property}_count in parent
  * [ ] order - make children create ordered or fix typer create date calc logic <---- can't do Typer "after" reply

Ex/ 38.1 holds lot's of access logs 38.11. Some lead to CSTOPs and need
to be recached. Flip will end up with 38.11-less 38.1's and distinct, 5000
clumped 38.11's in DataRF
"""
def flip(stationNo, typId, multipleProp, multipleTypId, multipleTypName, explicitInheritProps=[], mlimit=7500, useRF=False):

    startFlip = datetime.now()
    logLocn = re.sub(r'Data\/', 'Logs/', DATA_LOCN_TEMPL).format(stationNo)
    configLogging(stationNo, "FLIP")
    logging.info("Begin flipping {} of {}".format(typId, multipleProp))
    sourceDataLocn = DATARF_LOCN_TEMPL.format(stationNo) if useRF else DATA_LOCN_TEMPL.format(stationNo)
    dataReplyStore = FMQLReplyStore(sourceDataLocn)        
    destDataLocn = DATARF_LOCN_TEMPL.format(stationNo)
    rfReplyStore = FMQLReplyStore(destDataLocn, useZip=True)
    replyIter = dataReplyStore.iterator(onlyTypes=[typId])
    includeParentLabel = True if typId not in PATIENT_NAME_TYPES else False
    mIEN = 0 # Note: only works if walk all multiples so each gets unique no
    noMReplies = 0
    noMResources = 0
    mreply = {"results": [], "fmql": {"TYPE": multipleTypId, "AFTERIEN": str(mIEN), "LIMIT": str(mlimit)}}
    ptUpProp = "{}_container_".format(multipleProp)
    cntMultipleProp = "{}_count_".format(multipleProp) # adding in counter to parent for mult
    noResourcesWOMultiple = 0
    for i, reply in enumerate(replyIter, 1):
        replyFile = replyIter.currentReplyFile()
        reply["lessResults"] = []
        for resource in reply["results"]:
            if multipleProp not in resource:
                if "fmqlStopped" in resource and multipleProp in resource["fmqlStopped"]:
                    raise Exception("Didn't expect to run against stopped for multiple {} but resource {} of {} stopped".format(multipleProp, resource["_id"], replyFile))
                noResourcesWOMultiple += 1
                reply["lessResults"].append(resource) # still add
                continue # legitimate (unusual but ...)
            # container_ hierarchy properties in resource 
            # ie/ if resource in resource, pull in (ex/ 44_003 in 44_001 and want 44
            cprops = set(prop for prop in resource if re.search(r'_container_$', prop))
            for mresource in resource[multipleProp]:
                if ptUpProp in mresource:
                    raise Exception("'{}' in source data as a property - reserved".format(ptUpProp))
                if "fmqlStopped" in mresource: # simple for now
                    raise Exception("Don't expect STOP within multiple of {}".format(replyFile))
                mIEN += 1 # doing simple IENs from complete walk 
                mId = "{}-{}".format(multipleTypId, mIEN)
                mresource["_id"] = mId
                del mresource["ien"] # NOT KEEPING THIS
                mresource[ptUpProp] = {"id": resource["_id"]}
                if includeParentLabel and "label" in resource:
                    mresource[ptUpProp]["label"] = resource["label"]
                for rprop in resource:
                    if not (re.search(r'_container_', rprop) or rprop in explicitInheritProps):
                        continue
                    if rprop in mresource:
                        raise Exception("Name clash between prop {} of multiple and of those of its containers".format(rprop))
                    mresource[rprop] = resource[rprop]
                mreply["results"].append(mresource)
                noMResources += 1
                if len(mreply["results"]) == mlimit:
                    noMReplies += 1
                    mReplyFile = rfReplyStore.flush(mreply)
                    logging.info("MReply {} - {:,} w/{:,} flushed, total mresources now {:,}".format(mReplyFile, noMReplies, mlimit, noMResources))
                    mreply = {"results": [], "fmql": {"TYPE": multipleTypId, "AFTERIEN": str(mIEN), "LIMIT": str(mlimit)}}
            # logging.debug("Removing another processed {} of {:,} from parent".format(multipleProp, len(resource[multipleProp])))
            if cntMultipleProp in resource:
                raise Exception("counter for multiple prop {} is already in resource so can't reset - name clash".format(cntMultipleProp))
            resource[cntMultipleProp] = str(len(resource[multipleProp]))
            del resource[multipleProp] # will reflush w/o multiple prop
            reply["lessResults"].append(resource)
        # Flushing mprop-less source reply
        reply["results"] = reply["lessResults"]
        del reply["lessResults"]
        rfReplyStore.flush(reply)
        logging.info("Finished and reflushed source reply {} with its {:,} results but removed {} if present".format(replyFile, len(reply["results"]), multipleProp))
    if len(mreply["results"]) > 0:
        noMReplies += 1
        rfReplyStore.flush(mreply)
        logging.info("Last MReply of {:,} flushed".format(noMReplies)) 
    logging.info("Finished flipping in {} - {:,} replies processed and reflushed, {:,} multiple replies flushed with {:,} records, {:,} never had the multiple".format(str(datetime.now() - startFlip), i, noMReplies, noMResources, noResourcesWOMultiple))

"""
Basic Data Location setup - Cache Locations (Data x 3, Schema, Log)

VISTA_DATA_BASE_DIR/{stationNo}/...

3 Data Directories (DataV1, Data, DataRF (flipped/Reframe), Schema, Logs, TmpWorking
"""
def ensureCacheLocations(stationNo):
    baseLocation = VISTA_DATA_BASE_DIR
    if not os.path.isdir(baseLocation):
        raise Exception("Base Location {} doesn't exist".format(baseLocation))
    if not os.access(baseLocation, os.W_OK):
        raise Exception("No Write Permission for Base Location {}".format(baseLocation))
    cacheLocation = baseLocation + stationNo + "/"
    if not os.path.isdir(cacheLocation):
        os.mkdir(cacheLocation)
    for dataDir in ["DataV1", "Data", "DataRF"]:
        dloc = cacheLocation + dataDir + "/"
        if not os.path.isdir(dloc):
            os.mkdir(dloc)
    schemaLocation = cacheLocation + "Schema/"
    if not os.path.isdir(schemaLocation):
        os.mkdir(schemaLocation)
    logLocation = cacheLocation + "Logs/"
    if not os.path.isdir(logLocation):
        os.mkdir(logLocation)
    tmpLocation = cacheLocation + "TmpWorking/"
    if not os.path.isdir(tmpLocation):
        os.mkdir(tmpLocation)
    metaDataLocation = cacheLocation + "DataMeta/"
    if not os.path.isdir(metaDataLocation):
        os.mkdir(metaDataLocation)
           
"""
configure logging

ex/ logName="cacherLog"
""" 
def configLogging(stationNo, logName, logLevel="DEBUG"):

    log = logging.getLogger('')
    
    try: 
        mlogLevel = {"DEBUG": logging.DEBUG, "INFO": logging.INFO}[logLevel]
        log.setLevel(mlogLevel)
    except:
        log.setLevel(logging.DEBUG)

    logDir = LOG_LOCN_TEMPL.format(stationNo)
    fh = handlers.RotatingFileHandler("{}{}".format(logDir, "{}.out".format(logName)), maxBytes=(1048576*5), backupCount=7)
    format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    fh.setFormatter(format)
    log.addHandler(fh)

    ch = logging.StreamHandler(sys.stdout)
    format = logging.Formatter("%(levelname)s - %(message)s")
    ch.setFormatter(format)
    log.addHandler(ch)

# ############################# Test Driver ####################################

import os
import sys

def main():

    # zipAllRepliesByType("/data/vista/999/DataV1/")
    # return
    
    replyStore = FMQLReplyStore("/data/vista/999/Data/")
    flTyp = "80"
    iter = replyStore.iterator(onlyTypes=["80"])
    for flJSON in iter:
        print("First {}".format(len(flJSON["results"])))
        break
    lastReplyOfType = replyStore.lastReplyOfType("80")
    print("Last Reply of Type {}, {}".format(len(lastReplyOfType["results"]), lastReplyOfType["fmql"]))
    print(json.dumps(lastReplyOfType["results"][0], indent=4))
        
if __name__ == "__main__":
    main()
    



# FMQL Utilities (fmqlutils) 

import os
import re

"""
Base Directory for fmqlutils (Caching, Schema Typing, Reports) is VISTA_DATA_BASE_DIR.
By default it is '/data/vista/' but what happens if that location isn't set?

Specialize default base vista data location from shell or put into profile
    export FMQL_VISTA_DATA_BASE_DIR="/data/vistalocn/"

It is used with 
    from fmqlutils import VISTA_DATA_BASE_DIR
outside the module or
    from .. import VISTA_DATA_BASE_DIR 
from within
"""
VISTA_DATA_BASE_DIR = "/data/vista/" if os.environ.get("FMQL_VISTA_DATA_BASE_DIR") is None else re.sub(r'([^\/])$', '\\1/', os.environ.get("FMQL_VISTA_DATA_BASE_DIR"))
if not os.path.isdir(VISTA_DATA_BASE_DIR):
    raise Exception("fmqlutils module and submodules can't work - VISTA_DATA_BASE_DIR \"{}\" doesn't exist. Either _ln -s_ a real location to this directory OR set the environment variable \"FMQL_VISTA_DATA_BASE_DIR\" to the desired location".format(VISTA_DATA_BASE_DIR))




#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2017-2020 caregraf

import os
import sys
import re
import json
from collections import defaultdict, Counter
from datetime import datetime, date
import logging

from ..reporter.reportUtils import reportAbsAndPercent

from .reduceType import DEFAULT_MAX_NOMINAL_VALUES
    
# ################################## REPORT ##################################
                    
def reportReductions(typeDatas, outputFileName): # may get big
    try:
        os.remove(outputFileName) 
    except:
        pass
    if len(typeDatas) == 0:
        return # nothing to report
    flOut = open(outputFileName, "a")
    if len(typeDatas) == 1:
        flOut.write(reportReduction(typeDatas[0]))
        flOut.close()
        return
    allTypeDatas = [typeData for typeData in typeDatas if "_subTypeId" not in typeData]
    if len(allTypeDatas):
        if len(allTypeDatas) != 1:
            raise Exception("Expect one and only one ALL type data in a type data list")
        flOut.write("# All/Summary Report\n\n")
        flOut.write(reportReduction(allTypeDatas[0]))
    subTypeDatas = [typeData for typeData in typeDatas if "_subTypeId" in typeData]
    subTypeProp = subTypeDatas[0]["_subTypeId"].split(":")[0]
    flOut.write("\n\n# {:,} Sub Reports split by property \"{}\"\n\n".format(len(subTypeDatas), subTypeProp))
    for typeData in subTypeDatas:
        flOut.write(reportReduction(typeData))
    flOut.write("\n\n")
    flOut.close()
    
def reportReduction(typeData): # can use for a pass 2 as self contained

    def dtDelta(first, then): # Crude now - TODO more on timedelta, divmod and total_seconds 
        # delta = datetime.strptime(re.sub(r'Z$', '', then), "%Y-%m-%dT%H:%M:%S" ) - datetime.strptime(re.sub(r'Z$', '', first), "%Y-%m-%dT%H:%M:%S" )
        delta = int(then.split("-")[0]) - int(first.split("-")[0])
        return delta

    idMU = typeData["_id"] if "_label" not in typeData else "{} ({})".format(typeData["_id"], typeData["_label"])
    
    if "_subTypeId" in typeData:
        subTypeProp = typeData["_subTypeId"].split(":")[0]
        subTypeValue = typeData["_subTypeId"][len(subTypeProp) + 1:] # allow ':' prop val
        idMU += " property \"{}\" value \"{}\"".format(subTypeProp, subTypeValue)
        
    mu = "## {} Report\n\n".format(idMU)
    
    total = typeData["_total"]
    if "_numberFiltered" in typeData:
        filtered = typeData["_numberFiltered"]
        complete = total + filtered
        mu += "Total Selected: {}\n".format(reportAbsAndPercent(typeData["_total"], complete))
        mu += "Filtered: {}\n".format(reportAbsAndPercent(filtered, complete))
    else:
        mu += "Total: {:,}\n".format(total)
        
    if "_firstIEN" in typeData:
        mu += "First IEN: {}\n".format(typeData["_firstIEN"])  
              
    if "_createDateProp" in typeData:
        mu += "Create Date Property: {}\n".format(typeData["_createDateProp"])
        if typeData["_createDateProp"] in typeData:
            createDatePropInfo = typeData[typeData["_createDateProp"]]
            if "firstCreateDate" in createDatePropInfo:
                mu += "First Create Date: {}\n".format(createDatePropInfo["firstCreateDate"])
                mu += "Last Create Date: {}\n".format(createDatePropInfo["lastCreateDate"])
                delta = dtDelta(createDatePropInfo["firstCreateDate"], createDatePropInfo["lastCreateDate"])
                if delta != 0:
                    mu += "Span: {} years\n".format(delta)
                
    # prop that is the first or last date can vary
    if "_firstDateProps" in typeData:
        mu += "Order of Date Values (first/last dates):\n"
        mu += "\tFirsts:\n"
        for dp in sorted(typeData["_firstDateProps"], key=lambda x: typeData["_firstDateProps"][x], reverse=True):
            mu += "\t\t{} - {:,}\n".format(dp, typeData["_firstDateProps"][dp])  
        mu += "\tLasts:\n"
        for dp in sorted(typeData["_lastDateProps"], key=lambda x: typeData["_lastDateProps"][x], reverse=True): 
            mu += "\t\t{} - {:,}\n".format(dp, typeData["_lastDateProps"][dp])
            
    mu += "Properties:\n"
    atThreshold = 0 
    # .99999 (as using 'in between thresholds' is a bit of a kludge but ...) - can use in pass 2 reasoning for filter out edge edge
    thresholds = [1.0, 0.9999999999999, 0.99, 0.95, 0.9, 0.8, 0.5, 0.25, 0.1, 0.05]
    currentThresholdIndex = 0
    for j, prop in enumerate(sorted([key for key in typeData if not re.match(r'\_', key)], key=lambda x: typeData[x]["count"], reverse=True), 1):
    
        level = float(typeData[prop]["count"])/float(total)
        thresholdIndex = [i for i, t in enumerate(thresholds) if level <= t][-1] 
        if thresholdIndex != currentThresholdIndex:
            currentThresholdIndex = thresholdIndex
            mu += "\n\t------------- {}% cut off ------------------------\n".format(int(thresholds[thresholdIndex] * 100) if thresholdIndex > 1 else 100)
        propTypeMU = typeData[prop]["type"]
        
        propTimeSpanMU = ""
        if "firstCreateDate" in typeData[prop]:
            propTimeSpanMU = " - {} - ".format(typeData[prop]["firstCreateDate"].split("T")[0])
            firstYear = typeData[prop]["firstCreateDate"].split("-")[0]
            lastYear = typeData[prop]["lastCreateDate"].split("-")[0]
            if firstYear == lastYear:
                propTimeSpanMU += firstYear
            else:
                propTimeSpanMU += "{} --> {}".format(firstYear, lastYear)            
        
        mu += "\t{}. {} - {} - {:,} ({:.0%}){}{}\n".format(
            j, 
            prop, 
            propTypeMU, 
            typeData[prop]["count"], 
            float(typeData[prop]["count"])/float(total), 
            (" - " + ", ".join(typeData[prop]["rangeTypes"]) if "rangeTypes" in typeData[prop] else ""), 
            propTimeSpanMU
        )
        
        if "badLiteralValues" in typeData[prop]:
            mu += "\t\t** Bad Literal Values: {}\n".format(typeData[prop]["badLiteralValues"])
            
        # enum | pointer | date | list - order by key if int - otherwise by value
        if "byValueCount" in typeData[prop]:  
            for j, enumValue in enumerate(sorted(typeData[prop]["byValueCount"], key=lambda x: x if isinstance(x, int) else typeData[prop]["byValueCount"][x], reverse=True), 1):
                if j > DEFAULT_MAX_NOMINAL_VALUES:
                    mu += "\t\t... only showing top {} of {}\n".format(DEFAULT_MAX_NOMINAL_VALUES, len(typeData[prop]["byValueCount"]))
                    break
                mu += '\t\t{}. {} - {:,} ({:.1%})\n'.format(j, enumValue, typeData[prop]["byValueCount"][enumValue], float(typeData[prop]["byValueCount"][enumValue])/float(typeData[prop]["count"]))
        elif "rangeCount" in typeData[prop]: # big range pointer
            mu += "\t\tRange Count: {}\n".format(typeData[prop]["rangeCount"])
            
        if typeData[prop]["type"] == "LIST":
            FLIP_CANDIDATE_VARIETY = 50 # ie/ variability of size
            FLIP_CANDIDATE_SIZE = 10 # absolute size
            if "cstopped" in typeData[prop]: # only LIST
                mu += "\t\t**CSTOPed [FLIP CANDIDATE]: {:,} ({:.1%})\n".format(typeData[prop]["cstopped"], float(typeData[prop]["cstopped"])/float(typeData[prop]["count"]))
            elif len(typeData[prop]["byValueCount"]) > FLIP_CANDIDATE_VARIETY and sum(1 for listLen in typeData[prop]["byValueCount"] if int(listLen) > FLIP_CANDIDATE_SIZE):
                mu += "\t\t**FLIP CANDIDATE based on variety of lengths and size\n"
                
        if typeData[prop]["type"] == "DATE" and "byWeekDay" in typeData[prop]:
            wdMap = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
            mu += '\t\t------------------------------------------\n'
            for i, wd in enumerate(sorted(typeData[prop]["byWeekDay"]), 1): # from 0 Monday
                mu += '\t\t{}. {} - {}\n'.format(i, wdMap[wd], reportAbsAndPercent(typeData[prop]["byWeekDay"][wd], typeData[prop]["count"]))
                
    mu += "\n\n"
    return mu




#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2015-2020 caregraf

import re
import math
import textwrap

"""
TODO: consider combining into a report maker set ie/ make table (accept sets of rows too) and
make doc, add table, add graphic ref, add text, open section ...

Ala https://datapane.com/
"""
    
# #################### Markdown Utilities ###################

class MarkdownTable:

    """
    Supports:
    - first column # or not
    - explicit formatting by adding :'s to cols. Name "COL" given as :COL: is center aligned, :COL, left aligned, COL:, right aligned. Default is centered (:---:)
    - count of rows (for intro blurbs)
    """
    def __init__(self, cols, includeNo=True, defaultColMarker=":---:"):
        self.__noCols = len(cols)
        self.__includeNo = includeNo
        self.__rowNo = 0
        colNames = []
        colMarkers = []
        if includeNo:
            colNames.append("\#")
            colMarkers.append(":---:")
        for col in cols:
            formattedMatch = re.match(r'(:?)([^\:]+)(:?)$', col)
            if not formattedMatch: # may have :: or : inside etc => doesn't want format
                colNames.append(col)
                colMarkers.append(defaultColMarker)
                continue
            colNames.append(formattedMatch.group(2))
            if formattedMatch.group(1) == "" and formattedMatch.group(3) == "":
                colMarkers.append(defaultColMarker)
                continue
            colMarkers.append("{}---{}".format(formattedMatch.group(1), formattedMatch.group(3)))
        self.__tmu = " | ".join(colNames) + "\n" + " | ".join(colMarkers) + "\n"

    def addRow(self, vals):
        if len(vals) < self.__noCols:
            raise Exception("Need values for all cols")
        self.__rowNo += 1
        if self.__includeNo:
            rmu = "{}".format(str(self.__rowNo))
        else:
            rmu = ""
        for val in vals:
            if rmu:
                rmu += " | "
            if val == "":
                rmu += "&nbsp;"
            else:
                if isinstance(val, int):
                    rmu += format(val, ",")
                elif isinstance(val, list):
                    if len(val) != 2:
                        raise Exception("Use 2 piece list for url")
                    rmu += "[{}]({})".format(val[0], val[1])
                else:
                    rmu += str(val)
        rmu += "\n"
        self.__tmu += rmu

    def addSeparatorRow(self):
        rmu = " | ".join(["&nbsp;"] * (self.__noCols + (1 if self.__includeNo else 0))) + "\n"
        self.__tmu += rmu

    def md(self):
        return self.__tmu
        
    def rowCount(self):
        return self.__rowNo

"""
One line version
"""
def mdTable(rows, headers, count=False):
    """
    mdTable([["X", "Y"], [1]], ["X", "Y"])
    
    Decision NOT to force column values to be strings. Better to 
    format outside here (using escapeString etc) as approach differs
    with nature of data.
    """    
    noColumns = max(len(row) for row in rows)
    if len(headers) != noColumns:
        raise Exception("Expect headers to cover number of columns in all rows")
    if count:
        headers.insert(0, "\#")
    md = "\n" # ensure gap before
    # pad rows
    for i, row in enumerate(rows, 1):
        if len(row) < noColumns:
            for padNo in range(noColumns - len(row)):
                row.append("")
        if count:
            row.insert(0, str(i))
        if sum(1 for member in row if not isinstance(member, str)):
            raise Exception("Passed in a value that isn't a string")
    md += " | ".join(headers) + "\n"
    md += "|".join(["---" for i in range(len(headers))]) + "\n" 
    for i, row in enumerate(rows, 1):
        md += " | ".join(row) + "\n"
    return md
    
def mdHeaderRef(label):
    """
    Form: "#" with - in lowercase used by github for
    id's for headers. Use when want to refer to an anchor
    in a table of contents
    
    ex/ [" + label + "](#" + mdHeaderRef(label) + ")
    """
    return re.sub(' ', '-', label).lower()
  
def mdEscapeString(text, maxLen=-1):
    """
    - Using HTML entities to ensure no markdown interpretation of regexp or transforms
    - putting in ZERO WIDTH SPACE to tell browser when to split lines
      - removing \r first as deciding on new split (and \r will mess up MD
    """
    ESCAPES = {
        "(": "&#40;",
        ")": "&#41;",
        "[": "&#91;",
        "]": "&#93;",
        "?": "&#63;",
        "$": "&#38;",
        "<": "&#60;",
        ">": "&#62;",
        "^": "&#94;",
        "\\": "&#92;",
        "$": "&#36;",
        "|": "&#124;",
        "_": "&#95;"
    }
    
    text = re.sub(r'\r', ' ', text) # if embedded \r's get rid of them
    
    if maxLen != -1:
        text = "\r".join(textwrap.wrap(text, maxLen))
    #    # text = "\r".join([text[i:i+maxLen] for i in range(0, len(text), maxLen)])
    text = "".join(ESCAPES.get(c,c) for c in text)
    if maxLen != -1:
        text = re.sub(r'\r', "&#8203;", text)
    
    return text
    
# ################################## Sizing Utilities #######################

def convertSize(size):
   size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
   i = int(math.floor(math.log(size,1024)))
   p = math.pow(1024,i)
   s = round(size/p,2)
   if (s > 0):
       return '%s %s' % (s,size_name[i])
   else:
       return '0B'

# see: http://goo.gl/kTQMs
SYMBOLS = {
    'customary'     : ('B', 'KB', 'MB', 'GB', 'T', 'P', 'E', 'Z', 'Y'),
    # 'customary'     : ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'),
    'customary_ext' : ('byte', 'kilo', 'mega', 'giga', 'tera', 'peta', 'exa',
                       'zetta', 'iotta'),
    'iec'           : ('Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'),
    'iec_ext'       : ('byte', 'kibi', 'mebi', 'gibi', 'tebi', 'pebi', 'exbi',
                       'zebi', 'yobi'),
}    

def human2bytes(s):
    """
    Attempts to guess the string format based on default symbols
    set and return the corresponding bytes as an integer.
    When unable to recognize the format ValueError is raised.

      >>> human2bytes('0 B')
      0
      >>> human2bytes('1 K')
      1024
      >>> human2bytes('1 M')
      1048576
      >>> human2bytes('1 Gi')
      1073741824
      >>> human2bytes('1 tera')
      1099511627776

      >>> human2bytes('0.5kilo')
      512
      >>> human2bytes('0.1  byte')
      0
      >>> human2bytes('1 k')  # k is an alias for K
      1024
      >>> human2bytes('12 foo')
      Traceback (most recent call last):
          ...
      ValueError: can't interpret '12 foo'
    """
    init = s
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == '.':
        num += s[0]
        s = s[1:]
    num = float(num)
    letter = s.strip()
    for name, sset in SYMBOLS.items():
        if letter in sset:
            break
    else:
        if letter == 'k':
            # treat 'k' as an alias for 'K' as per: http://goo.gl/kTQMs
            sset = SYMBOLS['customary']
            letter = letter.upper()
        else:
            raise ValueError("can't interpret %r" % init)
    prefix = {sset[0]:1}
    for i, s in enumerate(sset[1:]):
        prefix[s] = 1 << (i+1)*10
    return int(num * prefix[letter])

# ##################### MUBVC ####################

"""
Note: forceShowCount is to allow an ST BVC use to force a count for
a single value if count < total
"""
def muBVC(bvc, separator=", ", forceShowCount=False, countOnlyIfOver=-1, totalOver=-1):
    def labelKey(k):
        if re.search(r' \[', k):
            return k.split(" [")[0]
        if not re.search(r':', k):
            return k
        return k.split(":")[1]
    if len(bvc) == 1:
        k = list(bvc)[0]
        if not forceShowCount:
            return labelKey(k)
        return "{} [{:,}]".format(labelKey(k), bvc[k])
    if countOnlyIfOver != -1 and len(bvc) > countOnlyIfOver:
        return "{:,}".format(len(bvc))
    if totalOver != -1 and len(bvc) > totalOver:
        sBVs = sorted(list(bvc), key=lambda x: bvc[x], reverse=True)
        return "{} and {:,} Others [{:,}]".format(
            separator.join(["{} [{:,}]".format(labelKey(k), bvc[k]) for k in sBVs[0:totalOver]]),
            len(sBVs[totalOver:]),
            sum(bvc[k] for k in sBVs[totalOver:])
        )
    mu = separator.join(["{} [{:,}]".format(labelKey(k), bvc[k]) for k in sorted(bvc, key=lambda x: bvc[x], reverse=True)])
    return mu
    
# ###################### Simple % stuff for text reports #########

# TODO: change to muX

def reportAbsAndPercent(abs, total):
    perc = reportPercent(abs, total) 
    if perc != "0.0%":
        return "{:,}".format(abs) + " (" + perc + ")"
    return "{:,}".format(abs)

def reportPercent(piece, total, dashIfZero=False):
    if not total: # can't divide by 0
        return "0.0%" if dashIfZero == False else "-"
    return str(makePercent(piece, total)) + "%"

def makePercent(piece, total):
    return round((float(piece) * 100)/float(total), 2)
    
# ######################### Tests and Demos #########################

def main():
    
    open("dump.md", "w").write(mdTable([["X", "(Y)"], [1], ["jessy went to the farm with mary to pick hay", "and pick up some stones and build a castle and call out to the birds before grabbing an egg and devouring it raw"]], ["X", "Y"]))
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2015-2020 caregraf

import os
import re 
import sys
import json
from datetime import datetime
from collections import defaultdict
import logging

from fmqlutils.cacher.cacherUtils import LOG_LOCN_TEMPL

def reportTrackedFiles(stationNo):
    trackZeroFile = "{}trackZeroFiles.json".format(LOG_LOCN_TEMPL.format(stationNumber))
    trackZeroNPartial = json.load(open(trackZeroFile))
    print("Partials", json.dumps(trackZeroNPartial["partials"]))
    print()
    print("Zeros", json.dumps(trackZeroNPartial["zeros"]))
    print()
    
# ################################# DRIVER #######################

def main():

    assert sys.version_info >= (3, 6)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")
        
    reportTrackedFiles(stationNo)



#!/usr/bin/env python

#
# (c) 2015-2020 caregraf
#

import sys
import urllib.request
import urllib.parse
import io
import ssl
import json
import re
import logging

from rpcutils.brokerRPC import VistARPCConnection, RPCLogger

"""
FMQL Endpoint Indirection - broker or REST and REST can be CSP (which has quirky errors)

TODO: merge in CacheInterface (may distinguish 3 - CSP/REST/BrokerIF
"""
                                
class FMQLRESTIF:
    """
    Indirection between endpoint for query so can go to CSPs etc.
        
    TODO: catch 503 etc in here and return clean ERROR codes
    """
    def __init__(self, fmqlEP, epWord="fmql"):
        self.fmqlEP = fmqlEP
        self.epWord = epWord
        
    def __str__(self):
        return "ENDPOINT: " + self.fmqlEP
        
    def invokeQuery(self, query):
        # rem: using GET and not POST so no encode to ascii too 
        queryURL = "{}?{}".format(self.fmqlEP, urllib.parse.urlencode({self.epWord: query}))
        if re.match("https", queryURL):
            # ignore bad cert - happens on test systems.
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
            breply = urllib.request.urlopen(queryURL, context=ctx).read()
        else:
            breply = urllib.request.urlopen(queryURL).read() 
        sreply = breply.decode("utf-8")
        try:
            # Want to preserve order of keys (as FMQL does) 
            jreply = json.loads(sreply) 
            return jreply
        except:
            if re.search(r'FMQL\.CSP', self.fmqlEP): # CSP endpoint
                #
                # Error: <b>Cannot allocate a license</b><br>
                # ErrorNo: <b>5915</b><br>
                # CSP Page: <b>/csp/fmquery/FMQL.CSP</b><br>
                # Namespace: <b>CHCS</b><br>
                #
                if re.search(r'ErrorNo', sreply):
                    error = re.search(r'Error: \<b\>([^\<]+)', sreply).group(1)
                    errorNo = re.search(r'ErrorNo: \<b\>([^\<]+)', sreply).group(1)
                    logging.error("CSP Error Response for query {} - {} - {}".format(query, errorNo, error))
            last500 = sreply[-500:]
            if len(sreply) > 500:
                last500 = "[LAST 500] {}".format(last500)
            logging.error("Problem Response for query {} - {}".format(query, last500))
            raise
            
class FMQLBrokerIF:

    def __init__(self, hostname, port, access, verify, osehraVISTA=False):
        conn = VistARPCConnection(hostname, int(port), access, verify, "CG FMQL QP USER", RPCLogger(), useOSEHRACipher=osehraVISTA)
        conn.connect()
        self.__connection = conn
        self.__epDescr = hostname + ":" + str(port)
        
    def __str__(self):
        return "BROKER ENDPOINT: " + self.__epDescr
        
    def invokeQuery(self, query):
    
        reply = self.__connection.invokeRPC("CG FMQL QP", [query])

        try: 
            # preserve order as FMQL does
            jreply = json.loads(reply)
        except:
            logging.error("Can't make JSON for response to query {} - {}".format(query, reply))
            raise
        
        return jreply

class CacheObjectInterface:
    """
    Utility for talking to FMQL through a Cache Object Interface. Key feature is managing sessions. Cache uses cookies for session identification. Cache limits the number of sessions on a server so it is important to use and reuse the same session.

    If Cache runs out of sessions, it will issue Service Unavailable 503 errors.

    Follows FMQLInterface

    NOTE: though ported to Python3, not yet tested and TODO needs to merge with FMQLIF above or be a clean difference
    """
    def __init__(self, ep):
        # ex/ http://...../FMQL.csp
        self.ep = ep
        self.cookie = ""

    def invokeRPC(self, name, params):
        # to match broker if for now
        return self.invokeQuery(params[0]).read()

    def invokeQuery(self, query):
        queryDict = {"query": query}
        queryurl = self.ep + "?" + urllib.parse.urlencode(queryDict).encode('ascii')
        request = urllib.request.Request(queryurl)
        if self.cookie:
            request.add_header('cookie', self.cookie)
        try:
            response = urllib.request.urlopen(request)
        except urllib.error.URLError as e:
            # 503 "Service Unavailable": The server cannot process the request due to a high load
            raise
        # Always reset the cookie - may be a new one if session idle for > 15 minutes.
        # SET-COOKIE: CSPSESSIONID-SP-57772-UP-csp-fmquery-=0010000100002g3gWldo9l0000fAj6SQgextDm2AmskX7GxQ--; path=/csp/fmquery/;  httpOnly;
        self.cookie = response.info().getheader('Set-Cookie')
        return response

def main():

    assert sys.version_info >= (3, 3)
    
    fmqlIF = FMQLRESTIF("http://localhost:9100/fmqlEP")
    res = fmqlIF.invokeQuery("DESCRIBE 8989_5-1")
    print(res)
    print()

    host = "localhost"
    port = 9330
    access = "QLFM1234"
    verify = "QLFM1234!!"
    connection = VistARPCConnection(host, int(port), access, verify, "CG FMQL QP USER", RPCLogger())
    reply = connection.invokeRPC("CG FMQL QP", ["DESCRIBE 8989_5-1"])
    print(reply)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2015-2020 caregraf

import os
import re 
import sys
import json
from datetime import datetime
from collections import defaultdict
import logging

from .fmqlIF import FMQLRESTIF, FMQLBrokerIF
from .cacherUtils import FMQLReplyStore, cacheFileToStore, ensureCacheLocations, V1RecordToV2, configLogging, DATAV1_LOCN_TEMPL, DATA_LOCN_TEMPL, SCHEMA_LOCN_TEMPL, LOG_LOCN_TEMPL, DATAMETA_LOCN_TEMPL
from .EXPLICIT_LS_CONFIGS import DEFAULT_LIMIT, DEFAULT_CSTOP, EXPLICIT_LIMITs, EXPLICIT_CSTOPs, FLIPs

"""
Basic FileMan data and schema cacher using FMQL.

TODO: FMQLSchemaStore to parallel FMQLReplyStore (may become FMQLDataStore)

Background: FileMan doesn't have enough indexes to support analytics. To fully analyze
data must first be cached. It can then be processed in a different DB (ex/ Mongo) or just
from disk.

Cache to "/data/vista/{stationNumber}/Data/" which it will create if it is absent

To test:
- du -h /data/{SYSTEM}/Data and /Schema
- old 2 download has 'File 2, limit 500 : all 1032150 of file 2 in 4:49:22.820342'

Invoke: 

    > nohup python -u cacher.py config/nv.json 8989_3 > cache.out & and tail -f cache.out; monitor with ps -aux, kill with kill -9 {pid}

See comment in main below for exact argument options.

ForCount is useful

    > fmqlcacher nconfigs/nv.json FORCOUNT10-20 
    
getting files with a count between 10 and 20
"""

# ########################## Data #########################

def cacheData(config):

    start = datetime.now()
            
    systemId = config["stationNumber"] if "name" not in config else "{} [{}]".format(config["name"], config["stationNumber"])
    logging.info("Caching Data of VistA {} [PID {}]".format(systemId, os.getpid()))
    
    fmqlIF = makeFMQLIF(config)
    jreply = fmqlIF.invokeQuery("ABOUT")
    if "error" in jreply:
        locn = DATAV1_LOCN_TEMPL.format(config["stationNumber"])
        logging.info("Older FMQL - caching in {}".format(locn))
        isV1 = True
    else:
        locn = DATA_LOCN_TEMPL.format(config["stationNumber"])
        logging.info("Using FMQL {} - caching in {}".format(jreply["version"], locn))
        isV1 = False
    useZip = config["useZip"] if "useZip" in config else True
    fmqlReplyStore = FMQLReplyStore(locn, useZip=useZip, makeDir=False)
    
    """
    Tracking Cache State
    - zeros is key to avoid trying and retrying the empty
    - partials arises with FORCOUNT where actual count >> FM's #.
    - TODO: complete now is NOT filled in
    ... These actual #'s can also be used in reports to replace FM's estimates.
    """
    def migrateOlderZeroAndPartials(stationNumber, trackFilesCached): # tmp til all migrated
        trackZeroFile = "{}trackZeroFiles.json".format(LOG_LOCN_TEMPL.format(stationNumber))
        try:
            zeroFiles = json.load(open(trackZeroFile))
        except:
            pass
        else:
            trackFilesCached["zeros"] = zeroFiles
            os.remove(trackZeroFile)
        trackMaxedFile = "{}trackForCountMaxedFiles.json".format(LOG_LOCN_TEMPL.format(stationNumber))
        try:
            hitMaxFiles = json.load(open(trackMaxedFile))
        except:
            pass
        else:
            trackFilesCached["partials"] = hitMaxFiles 
            os.remove(trackMaxedFile)
    trackFilesCachedFile = "{}trackFilesCached.json".format(LOG_LOCN_TEMPL.format(config["stationNumber"]))
    try:
        trackFilesCached = json.load(open(trackFilesCachedFile))
    except: # note: not yet noting/counting completes
        trackFilesCached = {"zeros": [], "partials": {}}
        migrateOlderZeroAndPartials(config["stationNumber"], trackFilesCached)
    originalZeroFilesCount = len(trackFilesCached["zeros"])
    originalPartialFilesCount = len(trackFilesCached["partials"])
    
    excludeTypes = set(config.get("excluded", []))
    includeTypes = set(config.get("included", [])) 
    includeCountMax = config.get("includeCountMax", -1)
    includeCountMin = config.get("includeCountMin", -1)  

    #
    # Option of Count Order or Id Order
    #
    # ... may retire: ala FMQLReplyStore, support a Schema Store
    # 
    def filesFromSchema(fmqlEP, schemaCacheLocation, typeOrder=False):
        """
        From SELECT TYPES - if in cache just load and read. Otherwise cache and
        then load and read.
        """
        query = "SELECT TYPES TOPONLY" # doing TOPONLY and not POPONLY in case POP is wrong
        cacheFile = re.sub(r' ', "_", query) + ".json"
        try:
            reply = json.load(open(schemaCacheLocation + cacheFile))
        except Exception:
            logging.info("First time through - must (re)cache {} to {} ...".format(query, schemaCacheLocation))
            reply = fmqlEP.invokeQuery(query)
            json.dump(reply, open(schemaCacheLocation + cacheFile, "w"))
        if typeOrder: # used if included explicit - otherwise order biggest first
            oresults = sorted(reply["results"], key=lambda res: float(re.sub(r'\_', '.', res["number"])))
        else:      
            oresults = sorted(reply["results"], key=lambda res: int(res.get("count", "0")), reverse=True)
        filesInOrder = []
        for i, result in enumerate(oresults, 1):
            fileId = re.sub(r'\.', '_', result["number"])
            filesInOrder.append({"id": fileId, "count": int(result.get("count", "0"))})
        logging.info("Returning {} top files - not all will have data".format(len(filesInOrder)))
        return filesInOrder

    defaultLimit = config["defaultLimit"]
    defaultCStop = config["defaultCStop"]
    problem = False
    cacheUpdated = False
    # cache in type order if list is explicit - otherwise count order (will
    # apply to typeSizeGroup too
    typeOrder = True if len(includeTypes) else False 
    schemaCacheLocation = SCHEMA_LOCN_TEMPL.format(config["stationNumber"])
    for i, fli in enumerate(filesFromSchema(fmqlIF, schemaCacheLocation, typeOrder=typeOrder), 1):

        if fli["id"] in excludeTypes:
            logging.debug("{}. File {} expecting count of {} - EXCLUDED SO SKIPPING".format(i, fli["id"], fli["count"]))
            continue

        if fli["id"] in trackFilesCached["zeros"]:
            logging.debug("{}. File {} expecting count of {} - KNOWN ZERO SO SKIPPING".format(i, fli["id"], fli["count"]))
            continue

        if includeTypes and fli["id"] not in includeTypes:
            logging.debug("{}. File {} expecting count of {} - NOT IN EXPLICIT INCLUDE LIST SO SKIPPING".format(i, fli["id"], fli["count"]))
            continue
                    
        if not (includeCountMax == -1 or includeCountMin == -1):
            try:
                fliCount = int(fli["count"])
            except:
                pass
            else:
                if fli["count"] < 0:
                    logging.debug("{}. File {} has invalid count of {} which we are treating as zero".format(i, fli["id"], fli["count"])) # second check below handles this
                if fli["count"] > includeCountMax:
                    logging.debug("{}. File {} has expected count of {:,} which exceeds max count {:,} so skipping".format(i, fli["id"], fli["count"], includeCountMax)) 
                    continue
                if includeCountMin != 0 and fli["count"] < includeCountMin:            
                    logging.debug("{}. File {} has expected count of {:,} which is below min count {:,} so skipping".format(i, fli["id"], fli["count"], includeCountMin)) 
                    continue
                # don't keep caching if actually bigger than threshold despite FM count
                maxNumber = includeCountMax
        else:
            maxNumber = -1 # turn off
                            
        logging.info("{}. File {} expecting count of {}{}".format(i, fli["id"], fli["count"], " ** Note POP says 0 but trying anyhow" if fli["count"] == 0 else ""))
            
        if "explicitLimits" in config and fli["id"] in config["explicitLimits"]:
            limit = config["explicitLimits"][fli["id"]]
        else:
            limit = defaultLimit
        if isV1:
            limit = int(limit/2) # size is for FMQL V10 - ~ 2x V1

        if "explicitCStops" in config and fli["id"] in config["explicitCStops"]:
            cstop = config["explicitCStops"][fli["id"]]
        else:
            cstop = defaultCStop

        try: # may get license error - make sure save zero files first
            cacheResult = cacheFileToStore(fmqlIF, fmqlReplyStore, fli["id"], limit=limit, cstop=cstop, maxNumber=maxNumber)
            if cacheResult == "HITMAX":
                trackFilesCached["partials"][fli["id"]] = maxNumber
                cacheUpdated = True # not saving result of complete
            elif cacheResult != "COMPLETE": # NO_SUCH_FILE, EMPTY_FILE 
                # will include 'no such files' and 'zero files'
                trackFilesCached["zeros"].append(fli["id"])
                cacheUpdated = True
            # not (yet) tracking completes (# is available by walking cache)
        except Exception as e:
            logging.exception(e)
            problem = True
            break

    timeTaken = str(datetime.now() - start)
    
    zeroFileMU = ""
    partialFileMU = ""
    if cacheUpdated:
        json.dump(trackFilesCached, open(trackFilesCachedFile, "w"))
        if len(trackFilesCached["zeros"]) > originalZeroFilesCount:
            zeroFileMU = " - zero file count went from {:,} to {:,}".format(originalZeroFilesCount, len(trackFilesCached["zeros"]))
        if len(trackFilesCached["partials"]) > originalPartialFilesCount:
            partialFileMU = " - hit max file ('partials') count went from {:,} to {:,}".format(originalPartialFilesCount, len(trackFilesCached["partials"]))

    if problem:
        logging.info("Caching finished UNSUCCESSFULLY at {} and took {}".format(datetime.now(), timeTaken))
    else:
        logging.info("Caching finished successfully at {} and took {}{}{}".format(datetime.now(), timeTaken, zeroFileMU, partialFileMU))
        
# ####################### Schema #####################
    
def cacheSchemas(config):

    start = datetime.now()
    
    systemId = config["stationNumber"] if "name" not in config else "{} [{}]".format(config["name"], config["stationNumber"])
    logging.info("Caching Schema of VistA {} [PID {}]".format(systemId, os.getpid()))

    fmqlIF = makeFMQLIF(config)
    
    schemaCacheLocation = SCHEMA_LOCN_TEMPL.format(config["stationNumber"])
    query = "SELECT TYPES"
    try:
        jreply = json.load(open(schemaCacheLocation + "SELECT_TYPES.json"))
    except:
        logging.info("First time through - must (re)cache {} to {} ...".format(query, schemaCacheLocation))
        jreply = fmqlIF.invokeQuery(query)
        json.dump(jreply, open(schemaCacheLocation + "SELECT_TYPES.json", "w"))

    fileIds = [re.sub(r'\.', "_", result["number"]) for result in jreply["results"]]
    logging.info("Must cache schema of {} files".format(len(fileIds)))

    alreadyCached = [re.match(r'SCHEMA\_([^\.]+)', f).group(1) for f in os.listdir(schemaCacheLocation) if os.path.isfile(os.path.join(schemaCacheLocation, f)) and re.search('\.json$', f) and re.match("SCHEMA_", f)]
    for i, fileId in enumerate(fileIds, 1):
        if fileId in alreadyCached:
            logging.info("{}. Schema of {} Already Cached".format(i, fileId))
            continue
        logging.info("{}. Caching Schema of {}".format(i, fileId))
        query = "DESCRIBE TYPE " + fileId
        queryStart = datetime.now()
        jreply = fmqlIF.invokeQuery(query)
        json.dump(jreply, open(schemaCacheLocation + "SCHEMA_" + fileId + ".json", "w"))

    # Phase 2 - cache SELECT TYPE REFS
    alreadyCached = [re.match(r'REFS\_([^\.]+)', f).group(1) for f in os.listdir(schemaCacheLocation) if os.path.isfile(os.path.join(schemaCacheLocation, f)) and re.search('\.json$', f) and re.match("REFS\_", f)]
    for i, fileId in enumerate(fileIds, 1):
        if fileId in alreadyCached:
            logging.info("{}. Refs of {} Already Cached".format(i, fileId))
            continue
        logging.info("{}. Caching Refs of {}".format(i, fileId))
        query = "SELECT TYPE REFS " + fileId
        queryStart = datetime.now()
        jreply = fmqlIF.invokeQuery(query)
        json.dump(jreply, open(schemaCacheLocation + "REFS_" + fileId + ".json", "w"))
        
    timeTaken = str(datetime.now() - start)
    logging.info("Caching SCHEMA/REF finished successfully at {} for {} files and took {}".format(datetime.now(), len(fileIds), timeTaken))
    
# ######################## DataMeta ##########################
    
"""
Cache the Meta Data needed to make a Meta Declaration for a VistA. Decoupled from
the creation of the meta record which is a separate function.

ADD?: may add other key files such as the code schemes, patient created first
etc. See what the following does first. This is ONLY meant to be cursory and
necessary (last sign on) data.
- LASTS:
  - http://localhost:9100/fmqlEP?fmql=DESCRIBE+LAST+757_01 --> last code added
  - last 3.081 or vital or ...
- parameters 8989_3 ... key global ones
- KSP: isProduction 
"""
def cacheDataMeta(config):
  
    fmqlIF = makeFMQLIF(config)
    fmqlReplyStore = FMQLReplyStore(DATAMETA_LOCN_TEMPL.format(config["stationNumber"]))
    v1ToV2 = V1RecordToV2(forMongo=True)

    if fmqlReplyStore.lastReplyOfType("8989_3") == None:
        cacheFileToStore(fmqlIF, fmqlReplyStore, "8989_3", 1, 0, 1)
        kspReply = fmqlReplyStore.lastReplyOfType("8989_3")
        kspResource = kspReply["results"][0]
        if "_id" not in kspResource:
            kspResource = v1ToV2.transform(kspResource) 
            kspReply["results"] = [kspResource]
            fmqlReplyStore.flush(kspReply)
        logging.info("Cached KSP for Meta for first time")
    else:
        logging.info("KSP Already Cached")
    
    # Using for 'last sign on' as PostMaster is daily and for first user
    # as always the first created
    if fmqlReplyStore.lastReplyOfType("200") == None:
        cacheFileToStore(fmqlIF, fmqlReplyStore, "200", 1, 0, 1) 
        postMasterReply = fmqlReplyStore.lastReplyOfType("200")
        postMasterResource = postMasterReply["results"][0]
        if "_id" not in postMasterResource:
            postMasterResource = v1ToV2.transform(postMasterResource)
            if postMasterResource["_id"] != "200-.5":
                raise Exception("cacheDataMeta assumption wrong - .5 is not first 200")
            postMasterReply["results"] = [postMasterResource]
            fmqlReplyStore.flush(postMasterReply)
        logging.info("Cached Post Master User for Meta for first time")
    else:
        logging.info("Post Master User Already Cached")
    
    # Always KERNEL - VIRGIN INSTALL 8.0 ?
    if fmqlReplyStore.lastReplyOfType("9_7") == None:
        cacheFileToStore(fmqlIF, fmqlReplyStore, "9_7", 1, 0, 1) 
        firstInstallReply = fmqlReplyStore.lastReplyOfType("9_7")
        firstInstallResource = firstInstallReply["results"][0]
        if "_id" not in firstInstallResource:
            firstInstallResource = v1ToV2.transform(firstInstallResource)
            firstInstallReply["results"] = [firstInstallResource]
            fmqlReplyStore.flush(firstInstallReply)
        logging.info("Cached First Install for Meta for first time")
    else:
        logging.info("First Install for Meta Already Cached")
    
# #################### FMQLIF wrapper (schema and data) #######################
    
def makeFMQLIF(config):
    # Should work for CSP and 'regular' REST endpoint
    if "fmqlEP" in config:
        fmqlIF = FMQLRESTIF(config["fmqlEP"], epWord=config.get("fmqlQuery", "fmql"))
        logging.info("Using REST Interface {}".format(config["fmqlEP"]))
    else:
        if "broker" not in config:
            raise Exception("Exiting - invalid 'config': neither 'broker' nor REST ('fmqlEP') settings available")
        # Must a/c for cypher in OSEHRA being different than Cypher in regular production VISTA (or maybe not if Vagrant VISTA is changed!)
        osehraVISTA = config["broker"].get("osehraVISTA", False) 
        fmqlIF = FMQLBrokerIF(config["broker"]["hostname"], config["broker"]["port"], config["broker"]["access"], config["broker"]["verify"], osehraVISTA=osehraVISTA)
        logging.info("Using RPC Broker Interface {}:{}".format(config["broker"]["hostname"], config["broker"]["port"]))
    return fmqlIF

# ############################# Driver ####################################

"""
./cacher.py {SYSTEM CONFIG FILE} META|SCHEMA|{TYP}|FORCOUNT{MIN\d+-MAX\d+}|{TYPE GROUP CONFIG FILE}

Always pass in "system config" and then can do one type or meta or schema or have a specific configuration override defaults and cover > one type.

- system config file is always needed to allow remote access. It defines
  - stationNumber (and [name])
  - end point/addressing parameters
  - [default cstop and limit settings for the system] (defaultCStop, defaultLimit)
  - [forcount min and max will cache types whose count (per the schema) fits into a range
    ... this is useful for caching smaller files (0-9, 10-49, 50-... etc. and note
    that FLIPs are excluded from such range caching to avoid problems.
  - [logLevel]
  - [explicitLimits, explicitCStops that override built in defaults for a specific system]
  ... it will NOT have 'included' as this goes in cross system type group files.
  
- type list config allows sets of types to be specified: typical example would be 
sets for certain report types. Format is
    {
        "name": "my group",
        "included": [... types
    }
  note that if a system doesn't have a file (per its schema), the cacher won't ask for it.
  This allows us to have generic type group files.
  
The role of EXPLICT_LS_SETTINGS: this sets type limit and cstop based on previous runs
against VistA. The default's may be overridden in the System Config file. Where there is
no setting for a type, the defaults are used.
"""
def main():

    assert sys.version_info >= (3, 4)
    
    try:
        sysConfigName = sys.argv[1].split(".")[0]
        arg2 = sys.argv[2]
    except IndexError:
        raise SystemExit("Usage _EXE_ {SYSTEM CONFIG FILE} META|SCHEMA|{TYP}|FORCOUNT{MIN-MAX}|explicit type list file}")
    
    if not os.path.isfile("{}.json".format(sysConfigName)):
        raise SystemExit("No system config file {}.json - exiting".format(sysConfigName))
    try:
        config = json.load(open("{}.json".format(sysConfigName)))
        if not ("stationNumber" in config and re.match(r'\d{3}$', config["stationNumber"])):
            raise SystemExit("Need three digit station number in config")
 
    except:
        raise SystemExit("Invalid system config {}.json - can't parse".format(sysConfigName))

    ensureCacheLocations(config["stationNumber"])
        
    configLogging(config["stationNumber"], "cacherLog", config["logLevel"] if "logLevel" in config else "INFO")
                
    if arg2 == "META":
        try:
            cacheDataMeta(config)
        except Exception as e:
            raise SystemExit("Can't cache meta: {}".format(str(e)))
        else:
            return
    
    if arg2 == "SCHEMA": 
        try:
            cacheSchemas(config)
        except Exception as e:
            raise SystemExit("Can't cache schema: {}".format(str(e)))
        else:
            return
            
    # Two constants - if no explicit setting then these rule
    if "defaultLimit" not in config:
        config["defaultLimit"] = DEFAULT_LIMIT
    if "defaultCStop" not in config:
        config["defaultCStop"] = DEFAULT_CSTOP 
            
    # Merging default explicit limits and superseeding with any set in system config
    if "explicitLimits" in config:
        EXPLICIT_LIMITs.update(config["explicitLimits"])
    config["explicitLimits"] = EXPLICIT_LIMITs
    if "explicitCStops" in config:
        EXPLICIT_CSTOPs.update(config["explicitCStops"])
    config["explicitCStops"] = EXPLICIT_CSTOPs
    
    if re.match(r'FORCOUNT', arg2):
        try:
            mtch = re.match(r'FORCOUNT(\d+)\-(\d+)$', arg2)
            config["includeCountMax"] = int(mtch.group(2))
            config["includeCountMin"] = int(mtch.group(1))
            # As a safety measure, FLIPs are excluded
            if "excluded" in config:
                config["excluded"].extend(FLIPs)
            else:
                config["excluded"] = FLIPs
        except:
            raise SystemExit("FORCOUNT Setting wrong")
    elif re.match(r'[\d\_]+$', arg2):
        typ = arg2
        config["included"] = [typ]
    else: # May override anything AND add an firstAFTERIENs
        typesConfigName = arg2.split(".")[0]
        if not os.path.isfile("{}.json".format(typesConfigName)):
            raise SystemExit("No types config file {}.json - exiting".format(typesConfigName))
        try:
            typesConfig = json.load(open("{}.json".format(typesConfigName)))
        except:
            raise SystemExit("Invalid types config {}.json - can't parse".format(typesConfigName))
        if "included" not in typesConfig:
            raise SystemExit("Expect 'included' to be set in types configuration files")
        config["included"] = typesConfig["included"]
        
    try:
        cacheData(config)
    except Exception as e:
        raise SystemExit("Can't cache data: {}".format(str(e)))

if __name__ == "__main__":
    main()




#!/usr/bin/env python
# -*- coding: utf8 -*-

# (c) 2017-2020 caregraf

import os
import sys
import re
import json
from collections import defaultdict, Counter
from datetime import datetime
import logging

from .. import VISTA_DATA_BASE_DIR
from ..cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator
from .reduceReportType import TYPER_REDUCTIONS_LOCN_TEMPL
from .reduceType import DEFAULT_MAX_NOMINAL_VALUES
from ..reporter.reportUtils import muBVC

"""
Expect 1 ALL (returned first) and many sub's with one or more properties. Note
that may have > that number of properties in split. 'expectSubTypeProperty' 
enforces the minimum.
"""
def splitTypeDatas(stationNo, typ, reductionLabel="", expectSubTypeProperties=[], expectSubTypeProperty="", expectNotEfficientWalked=False):
    if expectSubTypeProperty: # for backward compatibility
        expectSubTypeProperties = [expectSubTypeProperty]
    typeDatas = json.load(open(TYPER_REDUCTIONS_LOCN_TEMPL.format(stationNo) + "{}{}Reduction.json".format(typ, reductionLabel)))
    if isinstance(typeDatas, dict): # old and dictionary only of ALL
        return typeDatas, []
    allTypeDatas = [typeData for typeData in typeDatas if "_subTypeId" not in typeData]
    if len(allTypeDatas):
        if len(allTypeDatas) != 1:
            raise Exception("Expect one and only one ALL type data in a type data list")
    allTypeData = allTypeDatas[0]
    if expectNotEfficientWalked and "_efficientWalked" in allTypeData:
        raise Exception("Was efficient walked and shouldn't be")
    subTypeDatas = [typeData for typeData in typeDatas if "_subTypeId" in typeData]
    if len(subTypeDatas) and len(expectSubTypeProperties):
        subTypeProps = set(subTypeDatas[0]["_subTypeId"].split(":")[0].split("-"))
        if subTypeProps != set(expectSubTypeProperties):
            raise Exception("Expect sub type prop(s) \"{}\" but st has \"{}\"".format(":".join(sorted(expectSubTypeProperties)), subTypeDatas[0]["_subTypeId"]))
    elif expectSubTypeProperty:
        raise Exception("No sub type datas but expect {} as sub type property".format(expectSubTypeProperty))
    return allTypeData, subTypeDatas
    
"""
Combine sub types to broader criteria (including no criteria)

ex/ combineSubTypes(sts, ["prop1"]) where break is prop1, pr...
"""
def combineSubTypes(sts, subTypePropsSubset=[], forceCountProps=[], forceCountAllProps=False):
    def makeWantedSubTypeId(st, propsWanted):
        propsWanted = sorted(propsWanted)
        pieces = st["_subTypeId"].split(":")
        props = pieces[0].split("-")
        if len(set(propsWanted) - set(props)):
            raise Exception("Expect props wanted to be subset of those in st")
        allValues = [pieces[i] for i in range(1, len(pieces))]
        values = []
        for i, prop in enumerate(props):
            if prop in propsWanted:
                values.append(allValues[i])
        return "{}:{}".format("-".join(propsWanted), ":".join(values))
    stsBySTIdWanted = defaultdict(list)
    # ALL or broader groups
    if len(subTypePropsSubset):
        try:
            makeWantedSubTypeId(sts[0], subTypePropsSubset)
        except:
            raise Exception("Can't combine as subtype properties wanted aren't all in sts")
        for st in sts:
            stIdWanted = makeWantedSubTypeId(st, subTypePropsSubset)
            stsBySTIdWanted[stIdWanted].append(st)
    else:
        stsBySTIdWanted["ALL"] = sts
    createDateProp = "" if "_createDateProp" not in sts[0] else sts[0]["_createDateProp"]
    csts = []
    for stIdWanted in sorted(stsBySTIdWanted):
        # only prop kept  
        cst = {"_subTypeId": stIdWanted, "_total": 0}
        if createDateProp:
            cst["_createDateProp"] = createDateProp
        csts.append(cst)
        """
        For now, not taking 'rangeTypes', "firstCreateDate" or "range" and not
        the reduction of LIST ie/ just the counters and counts
        """
        for st in stsBySTIdWanted[stIdWanted]:
            cst["_total"] += st["_total"]
            for prop in st:
                if re.match(r'_', prop):
                    continue
                if prop not in cst:
                    cst[prop] = dict((field, st[prop][field]) for field in st[prop] if field in ["count", "type", "byValueCount", "byWeekDay", "firstCreateDate", "lastCreateDate"])
                    continue
                if "firstCreateDate" in cst[prop] and "lastCreateDate" in cst[prop]:
                    if "firstCreateDate" in st[prop] and "lastCreateDate" in st[prop]:
                        if st[prop]["firstCreateDate"] < cst[prop]["firstCreateDate"]:
                            cst[prop]["firstCreateDate"] = st[prop]["firstCreateDate"]
                        if st[prop]["lastCreateDate"] > cst[prop]["lastCreateDate"]:
                            cst[prop]["lastCreateDate"] = st[prop]["lastCreateDate"]                        
                    else: # may be overly cautious
                        del cst[prop]["firstCreateDate"]
                        del cst[prop]["lastCreateDate"]
                cst[prop]["count"] += st[prop]["count"]
                # merge counter only if in cst already; otherwise ignore. If not in st
                # then remove from cst ie/ either all with prop have cntType or no merge
                for cntType in ["byValueCount", "byWeekDay"]:
                    if cntType in cst[prop]:
                        if cntType in st[prop]: 
                            cst[prop][cntType] = Counter(cst[prop][cntType]) + Counter(st[prop][cntType])
                            # can get huge - ex IPs in user etc (typer does the same)
                            if forceCountAllProps == False and prop not in forceCountProps and len(cst[prop][cntType]) > DEFAULT_MAX_NOMINAL_VALUES:
                                del cst[prop][cntType]
                        else: # only delete from cst if another st with prop has none
                            del cst[prop][cntType]
    return csts
    
"""
Check Data: array of fileType and what's needed (array)
... ALL | TYPE | {REDUCTIONLABEL}

TODO/CONSIDER: -S variation ex fl = re.sub(r'\-0', '-S-0', fl) but consider
whether to just warn and force a STOP

Ex/ [{"fileType": "3_081", "check": "TYPE"}]
"""
def checkDataPresent(stationNumber, dataToCheck):
    hasAll = True
    for dataInfo in dataToCheck:
        ftyp = dataInfo["fileType"]
        if not isinstance(dataInfo["check"], list):
            dataInfo["check"] = [dataInfo["check"]]
        checked = {}
        for toCheck in dataInfo["check"]:
            if re.match("ALL", toCheck):
                if toCheck == "ALL":
                    # note: doesn't check completeness. One will do!
                    fl = "{}{}/Data/{}".format(VISTA_DATA_BASE_DIR, stationNumber, "{}-0.zip".format(ftyp))
                elif toCheck == "ALLRF":
                    fl = "{}{}/DataRF/{}".format(VISTA_DATA_BASE_DIR, stationNumber, "{}-0.zip".format(ftyp))
                else:
                    raise Exception("ALL options are ALL|ALLRF")
                if not os.path.isfile(fl):
                    checked[toCheck] = False
                    hasAll = False
                else:
                    checked[toCheck] = True
                continue
            reductionLabel = toCheck if toCheck != "TYPE" else ""
            fl = TYPER_REDUCTIONS_LOCN_TEMPL.format(stationNumber) + "{}{}Reduction.json".format(ftyp, reductionLabel)
            if not os.path.isfile(fl):
                checked[toCheck] = False
                hasAll = False
            else:
                checked[toCheck] = True
        del dataInfo["check"]
        dataInfo["checked"] = checked
    return hasAll, dataToCheck
    
"""
Convenience function for manipulating type reductions

Note: no need for singleValueCount as if singleValue => red[prop]["count"] has value
"""
def singleValue(typeRed, prop, ifMissingValue=""):
    if prop not in typeRed:
        if ifMissingValue:
            return ifMissingValue
        raise Exception("Unexpected missing property {} of reduction".format(prop))
    if not ("byValueCount" in typeRed[prop] and len(typeRed[prop]["byValueCount"]) == 1):
        raise Exception("Unexpected > 1 value for prop {} of reduction".format(prop))
    return list(typeRed[prop]["byValueCount"])[0]
    
"""
mu a bvc

Note: forceShowCount is to allow an ST BVC use to force a count for
a single value if count < total
"""
def muBVCOfSTProp(st, prop, separator=", ", addNotSet=True):
    if prop not in st:
        return "" # don't say NOTSET - just leave blank
    valueInfo = st[prop]
    if "byValueCount" not in valueInfo:
        raise Exception("Value missing BVC")
    bvc = valueInfo["byValueCount"]
    if addNotSet and st[prop]["count"] < st["_total"]:
        bvc["NOTSET"] = st["_total"] - st[prop]["count"]
    return muBVC(bvc, separator, st["_total"] != st[prop]["count"])
    
def thresholdProps(typeData):
    thresholds = [1.0, 0.99, 0.95, 0.9, 0.8, 0.5, 0.25, 0.1, 0.05]
    currentThresholdIndex = 0
    byThreshold = defaultdict(list)
    total = typeData["_total"]
    for j, prop in enumerate(sorted([key for key in typeData if not re.match(r'\_', key)], key=lambda x: typeData[x]["count"], reverse=True), 1):
        level = float(typeData[prop]["count"])/float(total)
        thresholdIndex = [i for i, t in enumerate(thresholds) if level <= t][-1]
        if thresholdIndex != currentThresholdIndex:
            currentThresholdIndex = thresholdIndex
        byThreshold[thresholds[currentThresholdIndex]].append(prop)
    return byThreshold

# ex/ refsOfST(st, "200") <=> userRefsOfST
def refsOfST(st, typ):
    refs = set()
    for prop in st:
        if re.match(r'\_', prop):
            continue
        if st[prop]["type"] == "LIST":
            refs |= refsOfST(st[prop]["reduction"], typ)
        if "rangeTypes" not in st[prop]:
            continue
        if typ not in st[prop]["rangeTypes"]:
            continue
        if "byValueCount" not in st[prop]:
            raise Exception("By Value Count not supported for {}/{}".format(st["_subTypeId"], prop))
        for ref in st[prop]["byValueCount"]:
            if not re.search(f'{typ}\-', ref):
                continue
            refs.add(ref)
    return refs 

"""
Time Series CSV for a type, using a day prop.

You can use upToDay (typically cutDate) but ONLY when dayProp is the type's creation property. timeBack is of form:
    {"months": #} or {"days": #} or {"years": #} or a combination ie/ > 1 arg

Specifically, header row and array of data rows as single string values
    "Date,ColId1,ColId2..."
    "2020-...,#,#"
    ...
where ColId1 ... breakdown by the Splitter. If there is no Splitter then
you see
    "Date,ALL"
    
Ex/ Splitter

    class Splitter:
        def __init__(self, stationNo):
            self.__stationNo = stationNo
        def split(self, resource):
            if ...
                return "" # suppress
            ... categorize
            return colId
    
Splitter can also SUPPRESS entries (ex/ if walking visits and only want encounters,
then can suppress the irrelevant)
    
Returns rows

Fits into a Pandas DF
    df = pd.read_csv(
        io.StringIO("\n".join(rows)),
        index_col=0, 
        parse_dates=True
    )
(not returning as easier to just serialize string for now)
    
To get colIds for display:
    rows[0].split(",")[1:]
"""        
def makeTypeTSCSVByDay(stationNo, typId, dayProp, upToDay="", timeBack=None, splitter=None):

    dataLocn = "{}{}/Data/".format(VISTA_DATA_BASE_DIR, stationNo) 
    if upToDay and timeBack:
        upToDayDT = datetime.strptime(upToDay.split("T")[0], "%Y-%m-%d")
        # timeBack == {"years": #} etc.
        onAndAfterDayDT = upToDayDT - relativedelta(**timeBack)
        onAndAfterDay = datetime.strftime(onAndAfterDayDT, "%Y-%m-%d")
        store = FMQLReplyStore(dataLocn)
        startAtReply = store.firstReplyFileOnOrAfterCreateDay(typId, dayProp, onAndAfterDay)
        print(f'TS CSV: iterating {typId} starting with reply {startAtReply} from {onAndAfterDay} to {upToDay} for creation property {dayProp}')
    else:
        onAndAfterDay = ""
        print(f'TS CSV: iterating {typId} from the start of data for property {dayProp}')
    resourceIter = FilteredResultIterator(dataLocn, typId, startAtReply=startAtReply)
    dayCounts = defaultdict(lambda: Counter())
    colIds = set()
    colId = "ALL"
    for resource in resourceIter:
        if dayProp not in resource:
            print(f'** Warning: missing {dayProp} from resource {resource["_id"]} - skipping')
            continue
        dt = resource[dayProp]["value"].split("T")[0]
        if onAndAfterDay:
            if dt < onAndAfterDay:
                continue
            if dt >= upToDay:
                break # assuming order
        if splitter:
            colId = splitter.split(resource)
            if colId == "": # can suppress
                continue
            colIds.add(colId)
        dayCounts[dt][colId] += 1
    colIds = sorted(list(colIds)) if len(colIds) else ["ALL"]
    rows = [f'Date,{",".join(colIds)}']
    for dt in sorted(dayCounts):
        row = [dt]
        for colId in colIds:
            cnt = 0 if colId not in dayCounts[dt] else dayCounts[dt][colId]
            row.append(str(cnt))
        rows.append(",".join(row))
    print(f'Returning TS with {len(colIds)} criteria over {len(rows)-1} dates from {rows[1].split(",")[0]} to {rows[-1].split(",")[0]}')
    return rows
